from typing import DefaultDict, Dict, Tuple

from array import ArrayType as array

import numpy as np

from ms_deisotope.data_source import ProcessedScan
from glycresoft.structure import FragmentCachingGlycopeptide


from glycresoft.tandem.spectrum_match.spectrum_match import ScanMatchManagingMixin


class PeakFragmentVector(ScanMatchManagingMixin):
    scan: ProcessedScan
    target: FragmentCachingGlycopeptide
    error_tolerance: float
    max_charge: int
    charge_vectors: DefaultDict[int, array]

    def __init__(self, scan: ProcessedScan, target: FragmentCachingGlycopeptide,
                 error_tolerance: float=2e-5, max_charge: int=2):
        super().__init__(scan, target)
        self.error_tolerance = error_tolerance
        self.max_charge = max_charge
        self.charge_vectors = DefaultDict(lambda: array('d'))
        self.build()

    def build_for_series(self, ion_series):
        zs = np.ones(self.max_charge)
        for frags in self.target.get_fragments(ion_series):
            for frag in frags:
                zs[:] = 0
                for peak in self.scan.deconvoluted_peak_set.all_peaks_for(frag.mass, self.error_tolerance):
                    z = abs(peak.charge)
                    if z > self.max_charge:
                        continue
                    zs[z - 1] = peak.intensity
                for z, i in enumerate(zs, 1):
                    self.charge_vectors[z].append(i)

    def build(self):
        if self.is_hcd():
            self.build_for_series('b')
            self.build_for_series('y')
        if self.is_exd():
            self.build_for_series('c')
            self.build_for_series('z')

    def flatten(self):
        acc = array('d')
        for k, v in self.charge_vectors.items():
            acc.extend(v)
        return np.asarray(acc)

    def correlate(self, other: 'PeakFragmentVector') -> float:
        xy = 0
        xx = 0
        yy = 0
        for z in range(1, self.max_charge + 1):
            x = np.asarray(self.charge_vectors[z])
            y = np.asarray(other.charge_vectors[z])
            xy += x.dot(y)
            xx += x.dot(x)
            yy += y.dot(y)
        return xy / (np.sqrt(xx) * np.sqrt(yy))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.target})"


class GlycanPeakFragmentVector(ScanMatchManagingMixin):
    scan: ProcessedScan
    target: FragmentCachingGlycopeptide
    error_tolerance: float
    max_charge: int
    fragment_map: Dict[Tuple[str, int], float]
    total: float

    def __init__(self, scan, target, error_tolerance: float=2e-5, max_charge: int = None):
        super().__init__(scan, target)
        if max_charge is None:
            max_charge = self.precursor_information.charge
        self.max_charge = max_charge
        self.error_tolerance = error_tolerance
        self.total = 0.0
        self.fragment_map = {}
        self.build()

    def build(self):
        total = 0.0
        for frag in self.target.stub_fragments(extended=True, extended_fucosylation=True):
            for peak in self.scan.deconvoluted_peak_set.all_peaks_for(frag.mass, self.error_tolerance):
                key = str(frag.glycosylation), peak.charge
                if key in self.fragment_map:
                    self.fragment_map[key] += peak.intensity
                else:
                    self.fragment_map[key] = peak.intensity
                total += peak.intensity
        self.total = total

    def correlate(self, other: 'GlycanPeakFragmentVector'):
        xy = 0
        xx = 0
        yy = 0
        keyspace = self.fragment_map.keys() | other.fragment_map.keys()
        for k in keyspace:
            x = self.fragment_map.get(k, 0)
            y = other.fragment_map.get(k, 0)
            xy += x * y
            xx += x ** 2
            yy += y ** 2
        return xy / (np.sqrt(xx) * np.sqrt(yy))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.target})"


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import correlate_fragment_map

    GlycanPeakFragmentVector.correlate = correlate_fragment_map
except ImportError:
    pass



import logging
import re


from collections import defaultdict


from glycresoft.serialize import (
    Peptide, Protein, DatabaseBoundOperation)

from .utils import slurp


from multiprocessing import Process, Queue, Event


logger = logging.getLogger("share_peptides")


class PeptideBlueprint(object):
    def __init__(self, base_peptide_sequence, modified_peptide_sequence, sequence_length,
                 calculated_mass, formula, count_glycosylation_sites, count_missed_cleavages,
                 count_variable_modifications, peptide_score, peptide_score_type, n_glycosylation_sites,
                 o_glycosylation_sites, gagylation_sites):
        self.base_peptide_sequence = base_peptide_sequence
        self.modified_peptide_sequence = modified_peptide_sequence
        self.sequence_length = sequence_length
        self.calculated_mass = calculated_mass
        self.formula = formula
        self.count_glycosylation_sites = count_glycosylation_sites
        self.count_missed_cleavages = count_missed_cleavages
        self.count_variable_modifications = count_variable_modifications
        self.peptide_score = peptide_score
        self.peptide_score_type = peptide_score_type
        self.n_glycosylation_sites = n_glycosylation_sites
        self.o_glycosylation_sites = o_glycosylation_sites
        self.gagylation_sites = gagylation_sites

    def as_db_instance(self):
        return Peptide(
            base_peptide_sequence=self.base_peptide_sequence, modified_peptide_sequence=self.modified_peptide_sequence,
            sequence_length=self.sequence_length,
            calculated_mass=self.calculated_mass, formula=self.formula,
            count_glycosylation_sites=self.count_glycosylation_sites,
            count_missed_cleavages=self.count_missed_cleavages,
            count_variable_modifications=self.count_variable_modifications, peptide_score=self.peptide_score,
            peptide_score_type=self.peptide_score_type, n_glycosylation_sites=self.n_glycosylation_sites,
            o_glycosylation_sites=self.o_glycosylation_sites,
            gagylation_sites=self.gagylation_sites)

    @classmethod
    def from_db_instance(cls, inst):
        return cls(
            base_peptide_sequence=inst.base_peptide_sequence, modified_peptide_sequence=inst.modified_peptide_sequence,
            sequence_length=inst.sequence_length,
            calculated_mass=inst.calculated_mass, formula=inst.formula,
            count_glycosylation_sites=inst.count_glycosylation_sites,
            count_missed_cleavages=inst.count_missed_cleavages,
            count_variable_modifications=inst.count_variable_modifications, peptide_score=inst.peptide_score,
            peptide_score_type=inst.peptide_score_type, n_glycosylation_sites=inst.n_glycosylation_sites,
            o_glycosylation_sites=inst.o_glycosylation_sites,
            gagylation_sites=inst.gagylation_sites)


class PeptideIndex(object):
    def __init__(self, store=None):
        if store is None:
            store = defaultdict(list)
        self.store = store

    def all_but(self, key):
        for bin_key, values in self.store.items():
            if key == bin_key:
                continue
            for v in values:
                yield v

    def populate(self, iterable):
        for peptide in iterable:
            self.store[peptide.protein_id].append(PeptideBlueprint.from_db_instance(peptide))


class PeptideSharer(DatabaseBoundOperation):
    def __init__(self, connection, hypothesis_id):
        DatabaseBoundOperation.__init__(self, connection)
        self.hypothesis_id = hypothesis_id
        self.index = PeptideIndex()
        self.index.populate(self._get_all_peptides())

    def find_contained_peptides(self, target_protein):
        session = self.session

        protein_id = target_protein.id

        i = 0
        j = 0
        target_protein_sequence = target_protein.protein_sequence
        keepers = []
        for peptide in self.stream_distinct_peptides(target_protein):

            match = self.decider_fn(peptide.base_peptide_sequence, target_protein_sequence)

            if match is not False:
                start, end, distance = match
                peptide = peptide.as_db_instance()
                peptide.hypothesis_id = self.hypothesis_id
                peptide.protein_id = protein_id
                peptide.start_position = start
                peptide.end_position = end
                keepers.append(peptide)
                j += 1
            i += 1
            if i % 100000 == 0:
                logger.info("%d peptides handled for %r", i, target_protein)
            if len(keepers) > 1000:
                session.bulk_save_objects(keepers)
                session.commit()
                keepers = []
                logger.info("%d peptides shared for %r", j, target_protein)

        session.bulk_save_objects(keepers)
        session.commit()
        # session.expunge_all()

    def decider_fn(self, query_seq, target_seq, **kwargs):
        match = re.search(query_seq, target_seq)
        if match:
            return match.start(), match.end(), 0
        return False

    def _get_all_peptides(self):
        return self.session.query(Peptide).filter(
            Peptide.hypothesis_id == self.hypothesis_id).all()

    def stream_distinct_peptides(self, protein):
        q = self.index.all_but(protein.id)

        for i in q:
            yield i


class PeptideSharingProcess(Process):
    def __init__(self, connection, hypothesis_id, input_queue, done_event=None):
        Process.__init__(self)
        self.connection = connection
        self.input_queue = input_queue
        self.hypothesis_id = hypothesis_id
        self.done_event = done_event

    def task(self):
        database = DatabaseBoundOperation(self.connection)
        session = database.session
        has_work = True

        sharer = PeptideSharer(self.connection, self.hypothesis_id)

        while has_work:
            try:
                work_items = self.input_queue.get(timeout=5)
                if work_items is None:
                    has_work = False
                    continue
            except Exception:
                if self.done_event.is_set():
                    has_work = False
                continue
            proteins = slurp(session, Protein, work_items, flatten=False)
            for protein in proteins:
                sharer.find_contained_peptides(protein)

    def run(self):
        self.task()


class MultipleProcessPeptideSharer(object):
    def __init__(self, connection, hypothesis_id, protein_ids, n_processes=4):
        self.connection = connection
        self.hypothesis_id = hypothesis_id
        self.protein_ids = protein_ids
        self.n_processes = n_processes

    def run(self):
        input_queue = Queue(100)
        done_event = Event()
        processes = [
            PeptideSharingProcess(
                self.connection, self.hypothesis_id, input_queue,
                done_event=done_event) for i in range(self.n_processes)
        ]
        protein_ids = self.protein_ids
        i = 0
        chunk_size = 20
        for process in processes:
            input_queue.put(protein_ids[i:(i + chunk_size)])
            i += chunk_size
            process.start()

        while i < len(protein_ids):
            input_queue.put(protein_ids[i:(i + chunk_size)])
            i += chunk_size

        done_event.set()
        for process in processes:
            process.join()



import pkg_resources

from glycresoft.database.builder.glycan import (
    GlycanCompositionEnzymeGraph, ExistingGraphGlycanHypothesisSerializer, GlycanTypes)
from glycresoft.database.prebuilt.utils import hypothesis_register, BuildBase

from io import StringIO


def load_graph():
    resource_bytes = pkg_resources.resource_string(
        __name__,
        "data/human_composition_enzyme_graph.json")
    resource_buffer = StringIO(resource_bytes.decode('utf8'))
    return GlycanCompositionEnzymeGraph.load(resource_buffer)


hypothesis_metadata = {
    "name": 'Biosynthesis Human N-Glycans',
    "hypothesis_type": 'glycan_composition',
    "description": "A collection of biosynthetically feasible *N*-glycans using enzymes commonly found in humans"
    " and limited to at most 26 monosaccharides. `GnTE` is explicitly omitted to make the space tractable."
}


@hypothesis_register(hypothesis_metadata['name'])
class BiosynthesisHumanNGlycansBuilder(BuildBase):

    def get_hypothesis_metadata(self):
        return hypothesis_metadata

    def build(self, database_connection, **kwargs):
        if kwargs.get('hypothesis_name') is None:
            kwargs['hypothesis_name'] = (self.hypothesis_metadata['name'])
        task = ExistingGraphGlycanHypothesisSerializer(
            load_graph(), database_connection, glycan_classification=GlycanTypes.n_glycan,
            **kwargs)
        task.start()
        return task



from csv import reader, writer
import os
import glob
import tempfile
import shutil

from glypy.utils import uid

from six import text_type, binary_type, PY3

from .spectrum_match import SpectrumMatch, SpectrumSolutionSet
from .ref import SpectrumReference, TargetReference


class TempFileManager(object):
    def __init__(self, base_directory=None):
        if base_directory is None:
            base_directory = tempfile.mkdtemp(prefix="glycresoft_tmp_")
        else:
            try:
                os.makedirs(base_directory)
            except OSError:
                if os.path.exists(base_directory) and os.path.isdir(base_directory):
                    pass
                else:
                    raise
        self.base_directory = base_directory
        self.cache = {}

    def get(self, key=None):
        if key is None:
            _key = ""
        else:
            _key = key
        if key in self.cache:
            return self.cache[key]
        name = "%s_%x" % (_key, uid())
        path = os.path.join(self.base_directory, name)
        self.cache[key] = path
        return path

    def clear(self):
        shutil.rmtree(self.base_directory)

    def dir(self, pattern='*'):
        return glob.glob(os.path.join(self.base_directory, pattern))

    def __repr__(self):
        return "TempFileManager(%r)" % self.base_directory


class FileWrapperBase(object):
    def flush(self):
        self.handle.flush()

    def close(self):
        self.handle.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class SpectrumSolutionSetWriter(FileWrapperBase):
    def __init__(self, sink, manager):
        self.manager = manager
        self.counter = 0
        self.sink = sink
        self._init_writer()

    def _init_writer(self):
        try:
            if PY3:
                self.handle = open(self.sink, 'wt', newline='')
            else:
                self.handle = open(self.sink, 'wb')
        except Exception:
            if hasattr(self.sink, 'write'):
                self.handle = self.sink
            else:
                raise
        self.writer = writer(self.handle)

    def write(self, solution_set):
        self.manager.put_scan(solution_set.scan)
        if PY3:
            values = [solution_set.scan.id]
            for spectrum_match in solution_set:
                self.manager.put_mass_shift(spectrum_match.mass_shift)
                for val in spectrum_match.pack():
                    if not isinstance(val, str):
                        val = str(val)
                    values.append(val)
            self.writer.writerow(values)
        else:
            values = [solution_set.scan.id]
            for spectrum_match in solution_set:
                self.manager.put_mass_shift(spectrum_match.mass_shift)
                values.extend(spectrum_match.pack())
            self.writer.writerow(list(map(bytes, values)))
        self.counter += 1

    def write_all(self, iterable):
        for item in iterable:
            self.write(item)
        self.flush()

    def extend(self, iterable):
        self.write_all(iterable)

    def append(self, solution_set):
        self.write(solution_set)

    def __len__(self):
        return self.counter


class SpectrumSolutionSetReader(FileWrapperBase):
    def __init__(self, source, manager, target_resolver=None, spectrum_match_type=SpectrumMatch):
        self.manager = manager
        self.source = source
        self.target_resolver = target_resolver
        self.spectrum_match_type = spectrum_match_type or SpectrumMatch
        self._init_reader()

    def _init_reader(self):
        try:
            if PY3:
                self.handle = open(self.source, 'rt')
            else:
                self.handle = open(self.source, 'rb')
        except Exception:
            if hasattr(self.source, 'read'):
                self.handle = self.source
            else:
                raise
        self.reader = reader(self.handle)

    def resolve_scan(self, scan_id):
        return self.manager.get_scan(scan_id)

    def resolve_target(self, target_id):
        if self.target_resolver is None:
            return self.manager.get_target(target_id)
        else:
            return self.target_resolver(target_id)

    def resolve_mass_shift(self, mass_shift_name):
        mass_shift = self.manager.get_mass_shift(mass_shift_name)
        return mass_shift

    def handle_line(self):
        row = next(self.reader)
        scan_id = str(row[0])
        spectrum_reference = self.resolve_scan(scan_id)
        i = 1
        n = len(row)
        members = []
        while i < n:
            match, i = self.spectrum_match_type.unpack(row, spectrum_reference, self, i)
            members.append(match)
        result = SpectrumSolutionSet(spectrum_reference, members)
        return result

    def __next__(self):
        return self.handle_line()

    next = __next__

    def __iter__(self):
        return self


class SpectrumSolutionResolver(object):
    def __init__(self):
        self.mass_shift_by_name = dict()
        self.target_cache = dict()
        self.scan_cache = dict()

    def clear(self):
        self.scan_cache.clear()
        self.target_cache.clear()

    def put_scan(self, scan):
        self.scan_cache[scan.id] = scan

    def get_scan(self, scan_id):
        try:
            return self.scan_cache[scan_id]
        except KeyError:
            return SpectrumReference(scan_id, None)

    def put_mass_shift(self, mass_shift):
        self.mass_shift_by_name[mass_shift.name] = mass_shift

    def get_mass_shift(self, name):
        return self.mass_shift_by_name[name]

    def get_target(self, id):
        try:
            return self.target_cache[id]
        except KeyError:
            ref = TargetReference(id)
            self.target_cache[id] = ref
            return ref


class SpectrumMatchStore(SpectrumSolutionResolver):
    _reader_type = SpectrumSolutionSetReader
    _writer_type = SpectrumSolutionSetWriter

    def __init__(self, tempfile_manager=None):
        if tempfile_manager is None:
            tempfile_manager = TempFileManager()
        self.temp_manager = tempfile_manager
        super(SpectrumMatchStore, self).__init__()

    def clear(self):
        self.temp_manager.clear()
        super(SpectrumMatchStore, self).clear()

    def writer(self, key):
        return self._writer_type(self.temp_manager.get(key), self)

    def reader(self, key, target_resolver=None):
        return self._reader_type(self.temp_manager.get(key), self, target_resolver=target_resolver)


class FileBackedSpectrumMatchCollection(object):
    def __init__(self, store, key, target_resolver=None):
        self.store = store
        self.key = key
        self.target_resolver = target_resolver
        self._size = None

    def __iter__(self):
        return self.store.reader(self.key, self.target_resolver)

    def __len__(self):
        if self._size is not None:
            return self._size
        i = 0
        for row in self:
            i += 1
        self._size = i
        return self._size



import unittest
import tempfile

from glycresoft.serialize.hypothesis.peptide import Peptide, Protein, Glycopeptide
from glycresoft.database.builder.glycopeptide import informed_glycopeptide
from glycresoft.database.builder.glycan import (
    CombinatorialGlycanHypothesisSerializer)
from glycresoft import serialize

from . import fixtures
from .test_constrained_combinatorics import FILE_SOURCE as GLYCAN_RULE_FILE_SOURCE


MZID_PATH = fixtures.get_test_data("AGP_Proteomics2.mzid")


class MzIdGlycopeptideTests(unittest.TestCase):

    def setup_tempfile(self, source):
        file_name = tempfile.mktemp() + '.tmp'
        open(file_name, 'w').write(source)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def test_build_hypothesis(self):
        glycan_file = self.setup_tempfile(GLYCAN_RULE_FILE_SOURCE)
        mzid_path = MZID_PATH
        db_file = glycan_file + '.db'

        glycan_builder = CombinatorialGlycanHypothesisSerializer(glycan_file, db_file)
        glycan_builder.run()

        glycopeptide_builder = informed_glycopeptide.MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer(
            mzid_path, db_file, glycan_builder.hypothesis_id)
        glycopeptide_builder.run()

        gp_count = glycopeptide_builder.query(Glycopeptide).count()
        with_uniprot = 769500
        with_uniprot_without_variable_signal_peptide = 659300
        without_uniprot = 651700
        without_any_external = 646000
        self.assertIn(
            gp_count, (with_uniprot, without_uniprot, without_any_external))

        redundancy = glycopeptide_builder.query(
            Glycopeptide.glycopeptide_sequence,
            Protein.name,
            serialize.func.count(Glycopeptide.glycopeptide_sequence)).join(
            Glycopeptide.protein).join(Glycopeptide.peptide).group_by(
                Glycopeptide.glycopeptide_sequence,
                Protein.name,
                Peptide.start_position,
                Peptide.end_position).yield_per(1000)

        for sequence, protein, count in redundancy:
            self.assertEqual(count, 1, "%s in %s has multiplicity %d" % (sequence, protein, count))

        for case in glycopeptide_builder.query(Glycopeptide).filter(
                Glycopeptide.glycopeptide_sequence ==
                "SVQEIQATFFYFTPN(N-Glycosylation)K{Hex:5; HexNAc:4; Neu5Ac:2}").all():
            self.assertAlmostEqual(case.calculated_mass, 4123.718954557139, 5)

        self.clear_file(glycan_file)
        self.clear_file(db_file)


if __name__ == '__main__':
    unittest.main()



import warnings
import struct
import logging

from typing import Any, ClassVar, List, Tuple, Type,  Union, Optional, TYPE_CHECKING, Generic, TypeVar
from enum import IntFlag

import glycopeptidepy

from glypy.utils import make_struct

from ms_deisotope import DeconvolutedPeakSet, isotopic_shift
from ms_deisotope.data_source import ProcessedScan
from ms_deisotope.data_source.metadata import activation
from glycresoft.chromatogram_tree.mass_shift import MassShiftBase

from glycresoft.structure import (
    ScanWrapperBase, ScanInformation, FragmentCachingGlycopeptide)

from glycresoft.structure.enums import SpectrumMatchClassification

from glycresoft.chromatogram_tree import Unmodified

from glycresoft.tandem.ref import TargetReference, SpectrumReference

if TYPE_CHECKING:
    from glycresoft.tandem.target_decoy.base import FDREstimatorBase
    from glycresoft.plotting.spectral_annotation import TidySpectrumMatchAnnotator
    from matplotlib.axes import Axes

neutron_offset = isotopic_shift()

LowCID = activation.dissociation_methods['low-energy collision-induced dissociation']

logger = logging.getLogger("glycresoft.spectrum_match")
logger.addHandler(logging.NullHandler())


class ScanMatchManagingMixin(ScanWrapperBase):
    __slots__ = ()

    target: Union[FragmentCachingGlycopeptide, Any]
    mass_shift: MassShiftBase

    def __init__(self, scan, target, mass_shift=None):
        if mass_shift is None:
            mass_shift = Unmodified
        self.scan = scan
        self.target = target
        self.mass_shift = mass_shift

    def drop_peaks(self):
        self.scan = ScanInformation.from_scan(self.scan)
        return self

    @staticmethod
    def load_peaks(scan) -> ProcessedScan:
        try:
            return scan.convert(fitted=False, deconvoluted=True)
        except AttributeError:
            return scan

    @staticmethod
    def threshold_peaks(deconvoluted_peak_set, threshold_fn=lambda peak: True):
        """
        Filter a deconvoluted peak set by a predicate function.

        Parameters
        ----------
        deconvoluted_peak_set : :class:`ms_deisotope.DeconvolutedPeakSet`
            The deconvoluted peaks to filter
        threshold_fn : Callable
            The predicate function to use to decide whether or not to keep a peak.

        Returns
        -------
        :class:`ms_deisotope.DeconvolutedPeakSet`
        """
        deconvoluted_peak_set = DeconvolutedPeakSet([
            p for p in deconvoluted_peak_set
            if threshold_fn(p)
        ])
        deconvoluted_peak_set._reindex()
        return deconvoluted_peak_set

    def is_hcd(self) -> bool:
        """
        Check whether the MSn spectrum was fragmented using a collisional dissociation
        mechanism or not.

        The result is cached on :attr:`scan`, so the interpretation does not need to be repeated.

        .. note:: If no activation information is present, the spectrum will be assumed to be HCD.

        Returns
        -------
        bool
        """
        scan = self.scan
        annotations = scan.annotations
        try:
            result = annotations['is_hcd']
        except KeyError:
            activation_info = scan.activation
            if activation_info is None:
                if scan.ms_level == 1:
                    result = False
                else:
                    warnings.warn(
                        "Activation information is missing. Assuming HCD")
                    result = True
            else:
                result = activation_info.has_dissociation_type(activation.HCD) or\
                    activation_info.has_dissociation_type(activation.CID) or\
                    activation_info.has_dissociation_type(LowCID) or\
                    activation_info.has_dissociation_type(
                        activation.UnknownDissociation)
            annotations['is_hcd'] = result
        return result

    def is_exd(self) -> bool:
        """
        Check if the scan was dissociated using an E*x*D method.

        This checks for ECD and ETD terms.

        This method caches its result in the scan's annotations.

        Returns
        -------
        bool

        See Also
        --------
        is_hcd
        """
        scan = self.scan
        annotations = scan.annotations
        try:
            result = annotations['is_exd']
        except KeyError:

            activation_info = scan.activation
            if activation_info is None:
                if scan.ms_level == 1:
                    result = False
                else:
                    warnings.warn(
                        "Activation information is missing. Assuming not ExD")
                    result = False
            else:
                result = activation_info.has_dissociation_type(activation.ETD) or\
                    activation_info.has_dissociation_type(activation.ECD)
            annotations['is_exd'] = result
        return result

    def mz_range(self) -> Tuple[float, float]:
        scan = self.scan
        annotations = scan.annotations
        try:
            result = annotations['mz_range']
        except KeyError:
            acquisition_info = scan.acquisition_information
            if acquisition_info is None:
                mz_range = (0, 1e6)
            else:
                lo = float('inf')
                hi = 0
                for event in acquisition_info:
                    for window in event:
                        lo = min(window.lower, lo)
                        hi = max(window.upper, hi)
                # No events/windows or an error
                if hi < lo:
                    mz_range = (0, 1e6)
                else:
                    mz_range = (lo, hi)
            annotations['mz_range'] = mz_range
            result = mz_range
        return result

    def get_auxiliary_data(self):
        return {}


class _SpectrumMatchBase(object):
    __slots__ = ['scan', 'target', "mass_shift"]

    scan: ProcessedScan
    target: Union[FragmentCachingGlycopeptide, Any]
    mass_shift: MassShiftBase


class SpectrumMatchBase(_SpectrumMatchBase, ScanMatchManagingMixin):
    """
    A base class for spectrum matches, a scored pairing between a structure
    and tandem mass spectrum.

    Attributes
    ----------
    scan: :class:`ms_deisotope.ProcessedScan`
        The processed MSn spectrum to match
    target: :class:`object`
        A structure that can be fragmented and scored against.
    mass_shift: :class:`~.MassShift`
        A mass shifting adduct that alters the precursor mass and optionally some
        of the fragment masses.

    """

    __slots__ = []

    scan: ProcessedScan
    target: Any
    mass_shift: MassShiftBase

    def _theoretical_mass(self) -> float:
        return self.target.total_composition().mass

    def precursor_mass_accuracy(self, offset: int=0) -> float:
        """
        Calculate the precursor mass accuracy in PPM, accounting for neutron offset
        and mass shift.

        Parameters
        ----------
        offset : int, optional
            The number of neutron errors to account for (the default is 0).

        Returns
        -------
        float:
            The precursor mass accuracy in PPM.
        """
        observed = self.precursor_ion_mass
        theoretical = self._theoretical_mass() + (
            offset * neutron_offset) + self.mass_shift.mass
        return (observed - theoretical) / theoretical

    def determine_precursor_offset(self,
                                   probing_range: int=3,
                                   include_error: bool=False) -> Union[int, Tuple[int, float]]:
        """
        Iteratively re-estimate what the actual offset to the precursor mass was.

        Parameters
        ----------
        probing_range : int, optional
            The range of neutron errors to account for. (the default is 3)
        include_error : bool, optional
            Whether or not to return both the offset and the estimated mass error in PPM.
            Defaults to False.

        Returns
        -------
        best_offset: int
            The best precursor offset.
        error: float
            The precursor mass error (in PPM) if `include_error` is :const:`True`
        """
        best_offset = 0
        best_error = float('inf')
        theoretical_mass_base = self._theoretical_mass() + self.mass_shift.mass
        observed = self.precursor_ion_mass
        for i in range(probing_range + 1):
            theoretical = theoretical_mass_base + i * neutron_offset
            error = abs((observed - theoretical) / theoretical)
            if error < best_error:
                best_error = error
                best_offset = i
        if include_error:
            return best_offset, best_error
        return best_offset

    def __reduce__(self):
        return self.__class__, (self.scan, self.target), self.__getstate__()

    def __getstate__(self):
        return {
            "mass_shift": self.mass_shift
        }

    def __setstate__(self, state):
        self.mass_shift = state.get("mass_shift", Unmodified)

    def get_top_solutions(self):
        return [self]

    def __eq__(self, other):
        try:
            target_id = self.target.id
        except AttributeError:
            target_id = None
        try:
            other_target_id = self.target.id
        except AttributeError:
            other_target_id = None
        return (self.scan == other.scan) and (self.target == other.target) and (
            target_id == other_target_id)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        try:
            target_id = self.target.id
        except AttributeError:
            target_id = None
        return hash((self.scan.id, self.target, target_id))


class SpectrumMatcherBase(SpectrumMatchBase):
    __slots__ = ["spectrum", "_score"]

    spectrum: DeconvolutedPeakSet
    _score: float

    def __init__(self, scan: ProcessedScan, target: Union[FragmentCachingGlycopeptide, Any],
                 mass_shift: Optional[MassShiftBase]=None):
        if mass_shift is None:
            mass_shift = Unmodified
        self.scan = scan
        self.spectrum = scan.deconvoluted_peak_set
        self.target = target
        self._score = 0
        self.mass_shift = mass_shift

    def drop_peaks(self):
        super(SpectrumMatcherBase, self).drop_peaks()
        self.spectrum = DeconvolutedPeakSet([])
        return self

    @property
    def score(self):
        """The aggregate spectrum match score.

        Returns
        -------
        float
        """
        return self._score

    @score.setter
    def score(self, value):
        self._score = value

    def match(self, *args, **kwargs):
        """Match theoretical fragments against experimental peaks.
        """
        raise NotImplementedError()

    def calculate_score(self, *args, **kwargs) -> float:
        """Calculate a score given the fragment-peak matching.

        This method should populate :attr:`_score`.

        Returns
        -------
        float
        """
        raise NotImplementedError()

    def base_peak(self) -> float:
        """Find the base peak intensity of the spectrum, the
        most intense peak's intensity.

        Returns
        -------
        float

        """
        try:
            return self.scan.annotations['_base_peak']
        except KeyError:
            peak = self.scan.base_peak()
            if peak is not None:
                value = self.spectrum.annotations['_base_peak'] = peak.intensity
            else:
                value = self.spectrum.annotations['_base_peak'] = 0.0
            return value

    @classmethod
    def evaluate(cls, scan, target, *args, **kwargs):
        """
        A high level method to construct a :class:`SpectrumMatcherBase`
        instance over a scan and target, call :meth:`match`, and
        :meth:`calculate_score`.

        Parameters
        ----------
        scan : :class:`~.ProcessedScan`
            The scan to match against.
        target : :class:`object`
            The structure to match.

        Returns
        -------
        :class:`SpectrumMatcherBase`
        """
        mass_shift = kwargs.pop("mass_shift", Unmodified)
        inst = cls(scan, target, mass_shift=mass_shift)
        inst.match(*args, **kwargs)
        inst.calculate_score(*args, **kwargs)
        return inst

    def __getstate__(self):
        state = super(SpectrumMatcherBase, self).__getstate__()
        state['score'] = self.score
        return state

    def __setstate__(self, state):
        super(SpectrumMatcherBase, self).__setstate__(state)
        self._score = state.get('score')

    def __reduce__(self):
        return self.__class__, (self.scan, self.target,), self.__getstate__()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.scan_id}, {self.spectrum}, {self.target}, {self.score})"

    def plot(self, ax: Optional['Axes']=None, **kwargs) -> 'TidySpectrumMatchAnnotator':
        """
        Plot the spectrum match, using the :class:`~.TidySpectrumMatchAnnotator`
        algorithm.

        Parameters
        ----------
        ax : :class:`matplotlib.Axis`, optional
            The axis to draw on. If not provided, a new figure
            and axis will be created.

        Returns
        -------
        :class:`~.TidySpectrumMatchAnnotator`
        """
        from glycresoft.plotting import spectral_annotation
        art = spectral_annotation.TidySpectrumMatchAnnotator(self, ax=ax)
        art.draw(**kwargs)
        return art

    @classmethod
    def get_score_set_type(cls) -> Type['ScoreSet']:
        return ScoreSet

    @classmethod
    def get_fdr_model_for_dimension(cls, label: str) -> Optional[Type['FDREstimatorBase']]:
        if label == 'peptide':
            from glycresoft.tandem.target_decoy import PeptideScoreSVMModel
            return PeptideScoreSVMModel

try:
    from glycresoft._c.tandem.tandem_scoring_helpers import base_peak
    SpectrumMatcherBase.base_peak = base_peak
except ImportError:
    pass


class MatchFlags(IntFlag):
    Unknown = 0
    Valid = 1
    Ambiguous = 2
    BestMatch = 4


F = TypeVar('F', bound=IntFlag)


class FlagProperty(Generic[F]):
    __slots__ = ('flag', )

    flag: F

    def __repr__(self):
        return f"{self.__class__.__name__}({self.flag})"

    def __init__(self, flag: F):
        self.flag = flag

    def __get__(self, obj, cls) -> bool:
        if obj is None:
            return self
        return bool(obj.flags & self.flag)

    def __set__(self, obj, value: bool):
        if value:
            obj.flags |= self.flag
        else:
            obj.flags &= ~self.flag


class SpectrumMatch(SpectrumMatchBase):
    """
    Represent a summarized spectrum match, which has been calculated already.

    Attributes
    ----------
    score: float
        The aggregate match score of the spectrum-structure pairing
    best_match: bool
        Whether or not this spectrum match is the best match for :attr:`scan`
    q_value: float
        The false discovery rate for the match
    id: int
        The unique identifier of the match

    """

    __slots__ = [
                 'score', '_best_match', 'data_bundle',
                 'q_value', 'id', 'valid', 'localizations',
                 'rank', 'cluster_id'
                ]

    score: float
    q_value: float

    valid: bool
    best_match: bool

    id: Optional[int]

    localizations: Optional[List['LocalizationScore']]
    data_bundle: Optional[Any]

    rank: Optional[int]
    cluster_id: Optional[int]

    def __init__(self, scan, target, score, best_match=False, data_bundle=None,
                 q_value=None, id=None, mass_shift=None, valid=True,
                 localizations=None, rank=0, cluster_id=None):
        super(SpectrumMatch, self).__init__(scan, target, mass_shift)

        self.score = score
        self._best_match = best_match
        self.data_bundle = data_bundle
        self.q_value = q_value
        self.id = id
        self.valid = valid
        self.localizations = localizations
        self.rank = rank
        self.cluster_id = cluster_id

    @property
    def best_match(self) -> bool:
        return self._best_match

    @best_match.setter
    def best_match(self, value: bool):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"... Setting {self.target}@{self.scan.id} best match to {value}",
                stacklevel=2
            )
        self._best_match = value

    @property
    def is_best_match(self) -> bool:
        return self.best_match

    @is_best_match.setter
    def is_best_match(self, value: bool):
        self.best_match = value

    def is_multiscore(self) -> bool:
        """Check whether this match has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return False

    def pack(self):
        """Package up the information of the match into the minimal information to
        describe the match.

        This method returns an opaque data carrier that can be reconstituted using
        :meth:`unpack`.

        Returns
        -------
        object
        """
        return (self.target.id, self.score, int(self.best_match), self.mass_shift.name)

    @classmethod
    def unpack(cls, data, spectrum, resolver, offset=0):
        """Reconstitute a :class:`SpectrumMatch` from packed
        data and a lookup table.

        Parameters
        ----------
        data : object
            The result of :meth:`pack`
        spectrum : :class:`~.ProcessedScan` or :class:`~.SpectrumReference`
            The spectrum that was matched.
        resolver : :class:`object`
            An object which can be used to resolve mass shifts and targets.
        offset : int, optional
            The offset into `data` to start reading from. Defaults to `0`.

        Returns
        -------
        match: :class:`SpectrumMatch`
            The reconstructed match
        offset: int
            The next offset to read from in `data`
        """
        i = offset
        target_id = int(data[i])
        score = float(data[i + 1])
        try:
            best_match = bool(int(data[i + 2]))
        except ValueError:
            best_match = bool(data[i + 2])
        mass_shift_name = data[i + 3]
        mass_shift = resolver.resolve_mass_shift(mass_shift_name)
        match = SpectrumMatch(
            spectrum,
            resolver.resolve_target(target_id),
            score,
            best_match,
            mass_shift=mass_shift)
        i += 4
        return match, i

    def clear_caches(self):
        """Clear caches on :attr:`target`.
        """
        try:
            self.target.clear_caches()
        except AttributeError:
            pass

    def __reduce__(self):
        return self.__class__, (self.scan, self.target, self.score, self.best_match,
                                self.data_bundle, self.q_value, self.id,
                                self.mass_shift, self.valid, self.localizations)

    def evaluate(self, scorer_type: Type[SpectrumMatcherBase], *args, **kwargs):
        """Re-evaluate this spectrum-structure pair.

        Parameters
        ----------
        scorer_type : :class:`SpectrumMatcherBase`
            The matcher type to use.

        Returns
        -------
        :class:`SpectrumMatcherBase`
        """
        if isinstance(self.scan, SpectrumReference):
            raise TypeError("Cannot evaluate a spectrum reference")
        elif isinstance(self.target, TargetReference):
            raise TypeError("Cannot evaluate a target reference")
        return scorer_type.evaluate(self.scan, self.target, *args, **kwargs)

    def __repr__(self):
        return "%s(%s, %s, %0.4f, %r)" % (
            self.__class__.__name__,
            self.scan, self.target, self.score, self.mass_shift)

    @classmethod
    def from_match_solution(cls, match: SpectrumMatcherBase) -> 'SpectrumMatch':
        """Create a :class:`SpectrumMatch` from another :class:`SpectrumMatcherBase`

        Parameters
        ----------
        match : :class:`SpectrumMatcherBase`
            The :class:`SpectrumMatcherBase` to convert

        Returns
        -------
        :class:`SpectrumMatch`
        """
        self = cls(match.scan, match.target, match.score, mass_shift=match.mass_shift)
        if hasattr(match, 'q_value'):
            self.q_value = match.q_value
        return self

    def clone(self):
        """Create a shallow copy of this object

        Returns
        -------
        :class:`SpectrumMatch`
        """
        return self.__class__(
            self.scan, self.target, self.score, self.best_match, self.data_bundle,
            self.q_value, self.id, self.mass_shift, self.localizations)

    def get_auxiliary_data(self):
        return self.data_bundle


class ModelTreeNode(object):
    def __init__(self, model, children=None):
        if children is None:
            children = {}
        self.children = children
        self.model = model

    def get_model_node_for(self, scan, target, *args, **kwargs):
        for decider, model_node in self.children.items():
            if decider(scan, target, *args, **kwargs):
                return model_node.get_model_node_for(scan, target, *args, **kwargs)
        return self

    def evaluate(self, scan, target, *args, **kwargs):
        node = self.get_model_node_for(scan, target, *args, **kwargs)
        return node.model.evaluate(scan, target, *args, **kwargs)

    def __call__(self, scan, target, *args, **kwargs):
        node = self.get_model_node_for(scan, target, *args, **kwargs)
        return node.model(scan, target, *args, **kwargs)

    def load_peaks(self, scan):
        return self.model.load_peaks(scan)


_ScoreSet = make_struct("ScoreSet", ['glycopeptide_score', 'peptide_score', 'glycan_score', 'glycan_coverage',
                                     "stub_glycopeptide_intensity_utilization",
                                     "oxonium_ion_intensity_utilization",
                                     "n_stub_glycopeptide_matches",
                                     "peptide_coverage", "total_signal_utilization"])


class ScoreSet(_ScoreSet):
    __slots__ = ()
    packer = struct.Struct("!ffffffffff")

    glycopeptide_score: float
    peptide_score: float
    glycan_score: float
    glycan_coverage: float
    stub_glycopeptide_intensity_utilization: float
    oxonium_ion_intensity_utilization: float
    n_stub_glycopeptide_matches: float
    peptide_coverag: float
    total_signal_utilization: float

    def __len__(self):
        return 4

    def __lt__(self, other):
        if self.glycopeptide_score < other.glycopeptide_score:
            return True
        elif abs(self.glycopeptide_score - other.glycopeptide_score) > 1e-3:
            return False

        if self.peptide_score < other.peptide_score:
            return True
        elif abs(self.peptide_score - other.peptide_score) > 1e-3:
            return False

        if self.glycan_score < other.glycan_score:
            return True
        elif abs(self.glycan_score - other.glycan_score) > 1e-3:
            return False

        if self.glycan_coverage < other.glycan_coverage:
            return True
        return False

    def __gt__(self, other):
        if self.glycopeptide_score > other.glycopeptide_score:
            return True
        elif abs(self.glycopeptide_score - other.glycopeptide_score) > 1e-3:
            return False

        if self.peptide_score > other.peptide_score:
            return True
        elif abs(self.peptide_score - other.peptide_score) > 1e-3:
            return False

        if self.glycan_score > other.glycan_score:
            return True
        elif abs(self.glycan_score - other.glycan_score) > 1e-3:
            return False

        if self.glycan_coverage > other.glycan_coverage:
            return True
        return False

    @classmethod
    def from_spectrum_matcher(cls, match):
        # Outdated, see Cython implementation
        return cls(match.score, match.peptide_score(), match.glycan_score(), match.glycan_coverage())

    def pack(self):
        return self.packer.pack(*self)

    @classmethod
    def unpack(cls, binary):
        return cls(*cls.packer.unpack(binary))

    @classmethod
    def field_names(cls):
        return [
            "total_score",
            "peptide_score",
            "glycan_score",
            "glycan_coverage",
            "stub_glycopeptide_intensity_utilization",
            "oxonium_ion_intensity_utilization",
            "n_stub_glycopeptide_matches",
            "peptide_coverage",
            "total_signal_utilization",
        ]

    def values(self):
        return [
            self.glycopeptide_score,
            self.peptide_score,
            self.glycan_score,
            self.glycan_coverage,
            self.stub_glycopeptide_intensity_utilization,
            self.oxonium_ion_intensity_utilization,
            self.n_stub_glycopeptide_matches,
            self.peptide_coverage,
            self.total_signal_utilization
        ]


class FDRSet(make_struct("FDRSet", ['total_q_value', 'peptide_q_value', 'glycan_q_value', 'glycopeptide_q_value'])):
    __slots__ = ()
    packer = struct.Struct("!ffff")

    total_q_value: float
    peptide_q_value: float
    glycan_q_value: float
    glycopeptide_q_value: float

    def pack(self):
        return self.packer.pack(*self)

    @classmethod
    def unpack(cls, binary):
        return cls(*cls.packer.unpack(binary))

    @classmethod
    def default(cls):
        return cls(1.0, 1.0, 1.0, 1.0)

    def __lt__(self, other):
        if self.total_q_value < other.total_q_value:
            return True
        elif abs(self.total_q_value - other.total_q_value) > 1e-3:
            return False

        if self.peptide_q_value < other.peptide_q_value:
            return True
        elif abs(self.peptide_q_value - other.peptide_q_value) > 1e-3:
            return False

        if self.glycan_q_value < other.glycan_q_value:
            return True
        elif abs(self.glycan_q_value - other.glycan_q_value) > 1e-3:
            return False

        if self.glycopeptide_q_value < other.glycopeptide_q_value:
            return True
        return False

    def __gt__(self, other):
        if self.total_q_value > other.total_q_value:
            return True
        elif abs(self.total_q_value - other.total_q_value) > 1e-3:
            return False

        if self.peptide_q_value > other.peptide_q_value:
            return True
        elif abs(self.peptide_q_value - other.peptide_q_value) > 1e-3:
            return False

        if self.glycan_q_value > other.glycan_q_value:
            return True
        elif abs(self.glycan_q_value - other.glycan_q_value) > 1e-3:
            return False

        if self.glycopeptide_q_value > other.glycopeptide_q_value:
            return True
        return False


try:
    _PyScoreSet = ScoreSet
    _PyFDRSet = FDRSet
    from glycresoft._c.tandem.spectrum_match import ScoreSet, FDRSet
    _has_c = True
except ImportError:
    _has_c = False


class MultiScoreSpectrumMatch(SpectrumMatch):
    __slots__ = ('score_set', 'match_type', '_q_value_set')

    score_set_type: ClassVar[Type] = ScoreSet

    def __init__(self, scan, target, score_set, best_match=False, data_bundle=None,
                 q_value_set=None, id=None, mass_shift=None, valid=True, match_type=None,
                 localizations=None, rank=0, cluster_id=None):
        if q_value_set is None:
            q_value_set = FDRSet.default()
        else:
            q_value_set = FDRSet(*q_value_set)
        self._q_value_set = None
        super(MultiScoreSpectrumMatch, self).__init__(
            scan, target, score_set[0], best_match, data_bundle, q_value_set[0],
            id, mass_shift, valid=valid, localizations=localizations, rank=rank,
            cluster_id=cluster_id)
        if isinstance(score_set, ScoreSet):
            self.score_set = score_set
        else:
            self.score_set = self.score_set_type(*score_set)
        self.q_value_set = q_value_set
        self.match_type = SpectrumMatchClassification[match_type]

    def is_multiscore(self):
        return True

    @property
    def q_value_set(self):
        return self._q_value_set

    @q_value_set.setter
    def q_value_set(self, value):
        self._q_value_set = value
        self.q_value = self._q_value_set.total_q_value

    def clone(self):
        """Create a shallow copy of this object

        Returns
        -------
        :class:`SpectrumMatch`
        """
        return self.__class__(
            self.scan, self.target, self.score_set, self.best_match, self.data_bundle,
            self.q_value_set, self.id, self.mass_shift, self.valid, self.match_type, self.localizations,
            self.rank, self.cluster_id)

    def __reduce__(self):
        return self.__class__, (self.scan, self.target, self.score_set, self.best_match,
                                self.data_bundle, self.q_value_set, self.id, self.mass_shift,
                                self.valid, self.match_type.value, self.localizations, self.rank,
                                self.cluster_id)

    def pack(self):
        return (self.target.id, self.score_set.pack(), int(self.best_match),
                self.mass_shift.name, self.match_type.value)

    @classmethod
    def from_match_solution(cls, match: SpectrumMatcherBase) -> 'MultiScoreSpectrumMatch':
        try:

            self = cls(
                match.scan,
                match.target,
                match.get_score_set_type().from_spectrum_matcher(match),
                mass_shift=match.mass_shift
            )
            if hasattr(match, 'q_value_set'):
                self.q_value_set = match.q_value_set
            return self
        except AttributeError:
            if isinstance(match, MultiScoreSpectrumMatch):
                return match
            else:
                raise


class LocalizationScore(object):
    __slots__ = ("position", "modification", "score")

    position: int
    modification: str
    score: float

    def __init__(self, position: int, modification: str, score: float):
        if isinstance(modification, (glycopeptidepy.Modification, glycopeptidepy.ModificationRule)):
            modification = modification.name
        elif not isinstance(modification, str):
            modification = str(modification)

        self.position = position
        self.modification = modification
        self.score = score

    def __reduce__(self):
        return self.__class__, (self.position, self.modification, self.score)

    def copy(self):
        return self.__class__(self.position, self.modification, self.score)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.position}, {self.modification}, {self.score})"

    def __str__(self):
        return f"{self.modification}:{self.position}:{self.score}"

    @classmethod
    def parse(cls, text: str):
        mod, pos, score = text.rsplit(":", 2)
        pos = int(pos)
        score = float(score)
        return cls(pos, mod, score)



import numpy as np

from ms_deisotope.scoring import g_test_scaled
from ms_deisotope.averagine import glycan, PROTON, mass_charge_ratio, TheoreticalIsotopicPattern
from ms_peak_picker.peak_set import FittedPeak
from brainpy import isotopic_variants

from glycresoft.chromatogram_tree import Unmodified

from .base import ScoringFeatureBase, epsilon


def envelope_to_peak_list(envelope):
    return [FittedPeak(e[0], e[1], 0, 0, 0, 0, 0, 0, 0) for e in envelope]


def scale_theoretical_isotopic_pattern(eid, tid):
    total = sum(p.intensity for p in eid)
    for p in tid:
        p.intensity *= total


def get_nearest_index(query_mz, peak_list):
    best_index = None
    best_error = float('inf')

    for i, peak in enumerate(peak_list):
        error = abs(peak.mz - query_mz)
        if error < best_error:
            best_error = error
            best_index = i
    return best_index


def align_peak_list(experimental, theoretical):
    retain = []
    for peak in experimental:
        retain.append(theoretical[get_nearest_index(peak.mz, theoretical)])
    # return TheoreticalIsotopicPattern(retain, theoretical[0].mz)
    return retain


def unspool_nodes(node):
    yield node
    for child in (node).children:
        for x in unspool_nodes(child):
            yield x


class IsotopicPatternConsistencyFitter(ScoringFeatureBase):
    feature_type = "isotopic_fit"

    def __init__(self, chromatogram, averagine=glycan, charge_carrier=PROTON, weighted=True):
        self.chromatogram = chromatogram
        self.averagine = averagine
        self.charge_carrier = charge_carrier
        self.scores = []
        self.intensity = []
        self.mean_fit = None
        self.weighted = weighted

        if chromatogram.composition is not None:
            if chromatogram.elemental_composition is not None:
                self.composition = chromatogram.elemental_composition
            else:
                raise Exception(chromatogram.composition)
        else:
            self.composition = None

        self.fit()

    def __repr__(self):
        return "IsotopicPatternConsistencyFitter(%s, %0.4f)" % (self.chromatogram, self.mean_fit)

    def generate_isotopic_pattern(self, charge, node_type=Unmodified):
        if self.composition is not None:
            tid = isotopic_variants(
                self.composition + node_type.composition,
                charge=charge, charge_carrier=self.charge_carrier)
            out = []
            total = 0.
            for p in tid:
                out.append(p)
                total += p.intensity
                if total >= 0.95:
                    break
            return out
        else:
            tid = self.averagine.isotopic_cluster(
                mass_charge_ratio(
                    self.chromatogram.neutral_mass + node_type.mass,
                    charge, charge_carrier=self.charge_carrier),
                charge,
                charge_carrier=self.charge_carrier)
            return tid

    def score_isotopic_pattern(self, deconvoluted_peak, node_type=Unmodified):
        eid, tid = self.prepare_isotopic_patterns(deconvoluted_peak, node_type)
        return g_test_scaled(None, eid, tid)

    def prepare_isotopic_patterns(self, deconvoluted_peak, node_type=Unmodified):
        tid = self.generate_isotopic_pattern(deconvoluted_peak.charge, node_type)
        eid = envelope_to_peak_list(deconvoluted_peak.envelope)
        scale_theoretical_isotopic_pattern(eid, tid)
        tid = align_peak_list(eid, tid)
        return eid, tid

    def fit(self):
        for node in self.chromatogram.nodes.unspool():
            for peak in node.members:
                score = self.score_isotopic_pattern(peak, node.node_type)
                self.scores.append(score)
                self.intensity.append(peak.intensity)
        self.intensity = np.array(self.intensity)
        self.scores = np.array(self.scores)
        if self.weighted:
            self.mean_fit = np.average(self.scores, weights=self.intensity)
        else:
            self.mean_fit = np.mean(self.scores)

    @classmethod
    def score(cls, chromatogram, *args, **kwargs):
        return max(1 - cls(chromatogram).mean_fit, epsilon)


class IsotopicPatternConsistencyModel(ScoringFeatureBase):
    feature_type = "isotopic_fit"

    def __init__(self, averagine=glycan, charge_carrier=PROTON):
        self.averagine = averagine
        self.charge_carrier = charge_carrier

    def fit(self, chromatogram):
        fitter = IsotopicPatternConsistencyFitter(
            chromatogram, averagine=self.averagine,
            charge_carrier=self.charge_carrier)
        return fitter

    def score(self, chromatogram, *args, **kwargs):
        return max(1 - self.fit(chromatogram).mean_fit, epsilon)



from glycresoft.trace.sink import ScanSink

from glycresoft.trace.match import (
    ChromatogramMatcher,
    GlycanChromatogramMatcher,
    GlycopeptideChromatogramMatcher)

from glycresoft.trace.extract import ChromatogramExtractor

from glycresoft.trace.evaluate import (
    ChromatogramEvaluator,
    LogitSumChromatogramEvaluator,
    LaplacianRegularizedChromatogramEvaluator)

from glycresoft.trace.process import (
    ChromatogramProcessor,
    LogitSumChromatogramProcessor,
    LaplacianRegularizedChromatogramProcessor,
    GlycopeptideChromatogramProcessor,
    NonSplittingChromatogramProcessor)


__all__ = [
    "ScanSink",
    "ChromatogramMatcher",
    "GlycanChromatogramMatcher",
    "GlycopeptideChromatogramMatcher",
    "ChromatogramExtractor",
    "ChromatogramEvaluator",
    "LogitSumChromatogramEvaluator",
    "LaplacianRegularizedChromatogramEvaluator",
    "ChromatogramProcessor",
    "LogitSumChromatogramProcessor",
    "LaplacianRegularizedChromatogramProcessor",
    "GlycopeptideChromatogramProcessor",
    "NonSplittingChromatogramProcessor"
]



cimport cython

from libc.math cimport pow, sqrt

cimport numpy as np

np.import_array()


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef (np.float32_t, np.float32_t) delta_distribution_parameters(np.float32_t[::1] apex_times, np.float32_t[::1] weights):
    cdef:
        Py_ssize_t i, n, size
        np.float32_t time, weight, acc, weight_acc, mean_time, time_stddev

    n = len(apex_times)
    with nogil:
        acc = 0
        weight_acc = 0
        size = 0
        for i in range(n):
            time = apex_times[i]
            weight = weights[i]
            acc += time * weight
            weight_acc += weight
        mean_time = acc / weight_acc
        acc = 0
        for i in range(n):
            time = apex_times[i]
            acc += pow(time - mean_time, 2) * weight
        time_stddev = sqrt(acc / (weight_acc - 1))
    return mean_time, time_stddev


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef (np.float32_t, np.float32_t, Py_ssize_t) filtered_delta_distribution_parameters(
                                                            np.float32_t[::1] apex_times,
                                                            np.float32_t[::1] weights,
                                                            cython.floating upper,
                                                            cython.floating lower):
    cdef:
        Py_ssize_t i, n, size
        np.float32_t time, weight, acc, weight_acc, mean_time, time_stddev

    n = len(apex_times)
    with nogil:
        acc = 0
        weight_acc = 0
        size = 0
        for i in range(n):
            time = apex_times[i]
            if time > lower and time < upper:
                weight = weights[i]
                acc += time * weight
                weight_acc += weight
                size += 1
        if size == 0:
            return 0, 1, 0
        mean_time = acc / weight_acc
        acc = 0
        for i in range(n):
            time = apex_times[i]
            if time > lower and time < upper:
                acc += pow(time - mean_time, 2) * weight
        time_stddev = sqrt(acc / (weight_acc - 1))
    return mean_time, time_stddev, size

@cython.boundscheck(False)
@cython.cdivision(True)
cpdef (double, double, double) posprocess_linear_fit(np.ndarray[np.float64_t, ndim=1, mode='c'] y,
                                                     np.float64_t[::1] residuals_,
                                                     np.float64_t[::1] w,
                                                     np.float64_t[::1] diag_H_):
    cdef:
        Py_ssize_t i, n, size
        np.float64_t[::1] y_
        double press
        double rss
        double tss
        double ybar
        double leave_one_out_err

    y_ = y
    ybar = y.mean()

    with nogil:
        press = 0.0
        tss = 0.0
        rss = 0.0
        n = y_.shape[0]

        for i in range(n):
            leave_one_out_err = (residuals_[i] / (1 - diag_H_[i]))
            press += w[i] * pow(leave_one_out_err, 2)
            rss += w[i] * pow(residuals_[i], 2)
            tss += w[i] * pow(y_[i] - ybar, 2)
    return (press, rss, tss)



import os

from glycresoft import serialize
from glycresoft.plotting import summaries, figax, SmoothingChromatogramArtist
from glycresoft.plotting.chromatogram_artist import ChargeSeparatingSmoothingChromatogramArtist
from glycresoft.scoring.utils import logit
from glycresoft.chromatogram_tree import ChromatogramFilter

from jinja2 import Template
from markupsafe import Markup


from glycresoft.output.report.base import (
    svguri_plot, ReportCreatorBase)


def chromatogram_figures(chroma):
    figures = []
    plot = SmoothingChromatogramArtist(
        [chroma], colorizer=lambda *a, **k: 'green', ax=figax()).draw(
        label_function=lambda *a, **k: "", legend=False).ax
    plot.set_title("Aggregated\nExtracted Ion Chromatogram", fontsize=24)
    chroma_svg = svguri_plot(
        plot, bbox_inches='tight', height=5, width=9, svg_width="100%")
    figures.append(chroma_svg)
    if len(chroma.mass_shifts) > 1:
        mass_shifts = list(chroma.mass_shifts)
        labels = {}
        rest = chroma
        for mass_shift in mass_shifts:
            with_mass_shift, rest = rest.bisect_mass_shift(mass_shift)
            labels[mass_shift] = with_mass_shift
        mass_shift_plot = SmoothingChromatogramArtist(
            labels.values(),
            colorizer=lambda *a, **k: 'green', ax=figax()).draw(
            label_function=lambda *a, **k: tuple(a[0].mass_shifts)[0].name,
            legend=False).ax
        mass_shift_plot.set_title(
            "mass_shift-Separated\nExtracted Ion Chromatogram", fontsize=24)
        mass_shift_separation = svguri_plot(
            mass_shift_plot, bbox_inches='tight', height=5, width=9, svg_width="100%")
        figures.append(mass_shift_separation)
    if len(chroma.charge_states) > 1:
        charge_separating_plot = ChargeSeparatingSmoothingChromatogramArtist(
            [chroma], ax=figax()).draw(
            label_function=lambda x, *a, **kw: str(
                tuple(x.charge_states)[0]), legend=False).ax
        charge_separating_plot.set_title(
            "Charge-Separated\nExtracted Ion Chromatogram", fontsize=24)
        charge_separation = svguri_plot(
            charge_separating_plot, bbox_inches='tight', height=5, width=9,
            svg_width="100%")
        figures.append(charge_separation)
    return figures


def chromatogram_link(chromatogram):
    id_string = str(chromatogram.id)
    return Markup("<a href=\"#detail-{0}\">{1}</a>").format(id_string, str(chromatogram.key))


class GlycanChromatogramReportCreator(ReportCreatorBase):
    def __init__(self, database_path, analysis_id, stream=None, threshold=5):
        super(GlycanChromatogramReportCreator, self).__init__(
            database_path, analysis_id, stream)
        self.set_template_loader(os.path.dirname(__file__))
        self.threshold = threshold
        self.glycan_chromatograms = ChromatogramFilter([])
        self.unidentified_chromatograms = ChromatogramFilter([])

    def glycan_link(self, key):
        match = self.glycan_chromatograms.find_key(key)
        if match is not None:
            return chromatogram_link(match)
        match = self.unidentified_chromatograms.find_key(key)
        if match is not None:
            return chromatogram_link(match)
        return None

    def prepare_environment(self):
        super(GlycanChromatogramReportCreator, self).prepare_environment()
        self.env.filters["logit"] = logit
        self.env.filters['chromatogram_figures'] = chromatogram_figures
        self.env.filters['glycan_link'] = self.glycan_link

    def make_template_stream(self):
        template_obj = self.env.get_template("overview.templ")

        ads = serialize.AnalysisDeserializer(
            self.database_connection._original_connection,
            analysis_id=self.analysis_id)

        self.glycan_chromatograms = gcs = ads.load_glycan_composition_chromatograms()
        # und = ads.load_unidentified_chromatograms()
        self.unidentified_chromatograms = und = ChromatogramFilter(
            ads.query(serialize.UnidentifiedChromatogram).filter(
                serialize.UnidentifiedChromatogram.analysis_id == self.analysis_id).all())

        if len(gcs) == 0:
            self.log("No glycan compositions were identified. Skipping report building")
            templ = Template('''
                <html>
                <style>
                body {
                    font-family: sans-serif;
                }
                </style>
                <body>
                    <h3>No glycan compositions were identified</h3>
                </body>
                </html>
                ''')
            return templ.stream()

        summary_plot = summaries.GlycanChromatographySummaryGraphBuilder(
            filter(lambda x: x.score > self.threshold, gcs + und))
        lcms_plot, composition_abundance_plot = summary_plot.draw(min_score=5)

        try:
            lcms_plot.ax.legend_.set_visible(False)
        except AttributeError:
            # The legend may not have been created
            pass
        lcms_plot.ax.set_title("Glycan Composition\nLC-MS Aggregated EICs", fontsize=24)

        fig = lcms_plot.ax.figure
        fig.set_figwidth(fig.get_figwidth() * 2.)
        fig.set_figheight(fig.get_figheight() * 2.)

        composition_abundance_plot.ax.set_title("Glycan Composition\nTotal Abundances", fontsize=24)
        composition_abundance_plot.ax.set_xlabel(
            composition_abundance_plot.ax.get_xlabel(), fontsize=14)

        def resolve_key(key):
            match = gcs.find_key(key)
            if match is None:
                match = und.find_key(key)
            return match

        template_stream = (template_obj.stream(
            analysis=ads.analysis, lcms_plot=svguri_plot(
                lcms_plot.ax, bbox_inches='tight', patchless=True,
                svg_width="100%"),
            composition_abundance_plot=svguri_plot(
                composition_abundance_plot.ax, bbox_inches='tight', patchless=True,
                svg_width="100%"),
            glycan_chromatograms=gcs,
            unidentified_chromatograms=und,
            resolve_key=resolve_key
        ))
        return template_stream



from collections import Counter


class ArithmeticMapping(Counter):
    def __init__(self, base=None, **kwargs):
        if base is not None:
            self.update(base)
        else:
            if kwargs:
                self.update(kwargs)

    def __missing__(self, key):
        return 0

    def __add__(self, other):
        inst = self.copy()
        for key, value in other.items():
            inst[key] += value
        return inst

    def __sub__(self, other):
        inst = self.copy()
        for key, value in other.items():
            inst[key] -= value
        return inst

    def __mul__(self, i):
        inst = self.copy()
        for key, value in self.items():
            inst[key] = value * i
        return inst

    def __imul__(self, i):
        for key, value in list(self.items()):
            self[key] = value * i
        return self

    def __div__(self, i):
        inst = self.copy()
        for key, value in self.items():
            inst[key] = value / i
        return inst

    def __idiv__(self, i):
        for key, value in list(self.items()):
            self[key] = value / i
        return self



from collections import OrderedDict
from itertools import product

import six

import numpy as np

from scipy.optimize import leastsq
from numpy import pi, sqrt, exp
from scipy.special import erf
from scipy.ndimage import gaussian_filter1d

from ms_peak_picker import search

from .base import ScoringFeatureBase, epsilon

try:
    trapezoid = np.trapz
except AttributeError:
    trapezoid = np.trapezoid

MIN_POINTS = 5
MAX_POINTS = 2000
SIGMA_EPSILON = 1e-3


def prepare_arrays_for_linear_fit(x, y):
    X = np.vstack((np.ones(len(x)), np.array(x))).T
    Y = np.array(y)
    return X, Y


def linear_regression_fit(x, y, prepare=False):
    if prepare:
        x, y = prepare_arrays_for_linear_fit(x, y)
    B = np.linalg.inv(x.T.dot(x)).dot(x.T.dot(y))
    Yhat = x.dot(B)
    return Yhat


def linear_regression_residuals(x, y):
    X, Y = prepare_arrays_for_linear_fit(x, y)
    Yhat = linear_regression_fit(X, Y)
    return (Y - Yhat) ** 2


def flat_line_residuals(y):
    residuals = (
        (y - ((y.max() + y.min()) / 2.))) ** 2
    return residuals


class PeakShapeModelBase(object):
    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    @classmethod
    def nargs(self):
        return six.get_function_code(self.shape).co_argcount - 1

    @classmethod
    def get_min_points(self):
        return getattr(self, "min_points", self.nargs() + 1)


def gaussian_shape(xs, center, amplitude, sigma):
    if sigma == 0:
        sigma = SIGMA_EPSILON
    norm = (amplitude) / (sigma * sqrt(2 * pi)) * \
        exp(-((xs - center) ** 2) / (2 * sigma ** 2))
    return norm


class GaussianModel(PeakShapeModelBase):
    min_points = 6

    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma = params
        return ys - GaussianModel.shape(xs, center, amplitude, sigma)

    @staticmethod
    def shape(xs, center, amplitude, sigma):
        return gaussian_shape(xs, center, amplitude, sigma)

    @staticmethod
    def guess(xs, ys):
        center = np.average(xs, weights=ys / ys.sum())
        height_at = np.abs(xs - center).argmin()
        apex = ys[height_at]
        sigma = np.abs(center - xs[[search.nearest_left(ys, apex / 2, height_at),
                                    search.nearest_right(ys, apex / 2, height_at + 1)]]).sum()
        return center, apex, sigma

    @staticmethod
    def params_to_dict(params):
        center, amplitude, sigma = params
        return OrderedDict((("center", center), ("amplitude", amplitude), ("sigma", sigma)))

    @staticmethod
    def center(params_dict):
        return params_dict['center']

    @staticmethod
    def spread(params_dict):
        return params_dict['sigma']


class SimplifiedGaussianModel(GaussianModel):
    def __init__(self, chromatogram):
        xs, ys = chromatogram.as_arrays()
        self.mean, self.base_amplitude, self.sigma = GaussianModel.guess(xs, ys)

    def guess(self, chromatogram):
        return (self.base_amplitude,)

    def shape(self, xs, amplitude):
        return GaussianModel.shape(xs, self.mean, amplitude, self.sigma)

    def fit(self, xs, ys, params):
        amplitude, = params
        return ys - self.shape(xs, amplitude)


def skewed_gaussian_shape(xs, center, amplitude, sigma, gamma):
    if sigma == 0:
        sigma = SIGMA_EPSILON
    amp_over_sigma_sqrt2pi = (amplitude) / (sigma * sqrt(2 * pi))
    sigma_squared_2 = 2 * sigma ** 2
    sigma_sqrt2 = (sigma * sqrt(2))

    norm = amp_over_sigma_sqrt2pi * \
        exp(-((xs - center) ** 2) / sigma_squared_2)
    skew = (1 + erf((gamma * (xs - center)) / sigma_sqrt2))
    return norm * skew


class SkewedGaussianModel(PeakShapeModelBase):
    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma, gamma = params
        return ys - SkewedGaussianModel.shape(xs, center, amplitude, sigma, gamma) * (
            sigma / 2. if abs(sigma) > 2 else 1.)

    @staticmethod
    def guess(xs, ys):
        center = np.average(xs, weights=ys / ys.sum())
        height_at = np.abs(xs - center).argmin()
        apex = ys[height_at]
        sigma = np.abs(center - xs[[search.nearest_left(ys, apex / 2, height_at),
                                    search.nearest_right(ys, apex / 2, height_at + 1)]]).sum()
        gamma = 1
        return center, apex, sigma, gamma

    @staticmethod
    def params_to_dict(params):
        center, amplitude, sigma, gamma = params
        return OrderedDict((("center", center), ("amplitude", amplitude), ("sigma", sigma), ("gamma", gamma)))

    @staticmethod
    def shape(xs, center, amplitude, sigma, gamma):
        return skewed_gaussian_shape(xs, center, amplitude, sigma, gamma)

    @staticmethod
    def center(params_dict):
        return params_dict['center']

    @staticmethod
    def spread(params_dict):
        return params_dict['sigma']


class PenalizedSkewedGaussianModel(SkewedGaussianModel):
    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma, gamma = params
        return ys - PenalizedSkewedGaussianModel.shape(xs, center, amplitude, sigma, gamma) * (
            sigma / 2. if abs(sigma) > 2 else 1.) * (gamma / 2. if abs(gamma) > 40 else 1.) * (
            center if center > xs[-1] or center < xs[0] else 1.)


def bigaussian_shape(xs, center, amplitude, sigma_left, sigma_right):
    if sigma_left == 0:
        sigma_left = SIGMA_EPSILON
    if sigma_right == 0:
        sigma_right = SIGMA_EPSILON
    ys = np.zeros_like(xs, dtype=np.float32)
    left_mask = xs < center
    ys[left_mask] = amplitude / sqrt(2 * pi) * np.exp(-(xs[left_mask] - center) ** 2 / (2 * sigma_left ** 2))
    right_mask = xs >= center
    ys[right_mask] = amplitude / sqrt(2 * pi) * np.exp(-(xs[right_mask] - center) ** 2 / (2 * sigma_right ** 2))
    return ys


class BiGaussianModel(PeakShapeModelBase):

    @staticmethod
    def center(params_dict):
        return params_dict['center']

    @staticmethod
    def spread(params_dict):
        return (params_dict['sigma_left'] + params_dict['sigma_right']) / 2.

    @staticmethod
    def shape(xs, center, amplitude, sigma_left, sigma_right):
        return bigaussian_shape(xs, center, amplitude, sigma_left, sigma_right)

    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma_left, sigma_right = params
        return ys - BiGaussianModel.shape(
            xs, center, amplitude, sigma_left, sigma_right) * (
            center if center > xs[-1] or center < xs[0] else 1.)

    @staticmethod
    def params_to_dict(params):
        center, amplitude, sigma_left, sigma_right = params
        return OrderedDict(
            (("center", center), ("amplitude", amplitude), ("sigma_left", sigma_left), ("sigma_right", sigma_right)))

    @staticmethod
    def guess(xs, ys):
        center = np.average(xs, weights=ys / ys.sum())
        height_at = np.abs(xs - center).argmin()
        apex = ys[height_at]
        sigma = np.abs(center - xs[[search.nearest_left(ys, apex / 2, height_at),
                                    search.nearest_right(ys, apex / 2, height_at + 1)]]).sum()
        return center, apex, sigma, sigma


class FittedPeakShape(object):
    def __init__(self, params, shape_model):
        self.params = params
        self.shape_model = shape_model

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def items(self):
        return self.params.items()

    def __iter__(self):
        return iter(self.params)

    def shape(self, xs):
        return self.shape_model.shape(xs, **self.params)

    def __getitem__(self, key):
        return self.params[key]

    def __repr__(self):
        return "Fitted{self.shape_model.__class__.__name__}({params})".format(
            self=self, params=", ".join("%s=%0.3f" % (k, v) for k, v in self.params.items()))

    @property
    def center(self):
        return self['center']

    @property
    def amplitude(self):
        return self['amplitude']

    @property
    def spread(self):
        return self.shape_model.spread(self)


class ChromatogramShapeFitterBase(ScoringFeatureBase):
    feature_type = "line_score"

    def __init__(self, chromatogram, smooth=True, fitter=PenalizedSkewedGaussianModel()):
        self.chromatogram = chromatogram
        self.smooth = smooth
        self.xs = None
        self.ys = None
        self.line_test = None
        self.off_center = None
        self.shape_fitter = fitter

    def is_invalid(self):
        n = len(self.chromatogram)
        return n < MIN_POINTS or n < self.shape_fitter.get_min_points()

    def handle_invalid(self):
        self.line_test = 1 - 5e-6

    def extract_arrays(self):
        self.xs, self.ys = self.chromatogram.as_arrays()
        if self.smooth:
            self.ys = gaussian_filter1d(self.ys, self.smooth)
        if len(self.xs) > MAX_POINTS:
            new_xs = np.linspace(self.xs.min(), self.xs.max(), MAX_POINTS)
            new_ys = np.interp(new_xs, self.xs, self.ys)
            self.xs = new_xs
            self.ys = new_ys
            self.ys = gaussian_filter1d(self.ys, self.smooth)

    def compute_residuals(self):
        raise NotImplementedError()

    def compute_fitted(self):
        raise NotImplementedError()

    def null_model_residuals(self):
        residuals = linear_regression_residuals(self.xs, self.ys)
        return residuals

    def perform_line_test(self):
        residuals = self.compute_residuals()
        line_test = (residuals ** 2).sum() / (
            (self.null_model_residuals()).sum())
        self.line_test = max(line_test, 1e-5)

    def plot(self, ax=None):
        if ax is None:
            from matplotlib import pyplot as plt
            fig, ax = plt.subplots(1)
        ob1 = ax.plot(self.xs, self.ys, label='Observed')[0]
        ob2 = ax.scatter(self.xs, self.ys, label='Observed')
        f1 = ax.plot(self.xs, self.compute_fitted(), label='Fitted')[0]
        r1 = ax.plot(self.xs, self.compute_residuals(), label='Residuals')[0]
        ax.legend(
            (
                (ob1, ob2),
                (f1,),
                (r1,)
            ), ("Observed", "Fitted", "Residuals")
        )
        return ax

    @property
    def fit_parameters(self):
        raise NotImplementedError()

    @classmethod
    def score(cls, chromatogram, *args, **kwargs):
        return max(1 - cls(chromatogram).line_test, epsilon)

    def integrate(self):
        time = self.xs
        smoothed = self.compute_fitted()
        return trapezoid(smoothed, time)


class ChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, smooth=True, fitter=PenalizedSkewedGaussianModel()):
        super(ChromatogramShapeFitter, self).__init__(chromatogram, smooth=smooth, fitter=fitter)

        self.params = None
        self.params_dict = None

        if self.is_invalid():
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()
            self.off_center_factor()

    @property
    def fit_parameters(self):
        return self.params_dict

    def __repr__(self):
        return "ChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def off_center_factor(self):
        center = self.shape_fitter.center(self.params_dict)
        spread = self.shape_fitter.spread(self.params_dict)
        self.off_center = abs(1 - abs(1 - (2 * abs(
            self.xs[self.ys.argmax()] - center) / abs(spread))))
        if self.off_center > 1:
            self.off_center = 1. / self.off_center
        self.line_test /= self.off_center

    def compute_residuals(self):
        return self.shape_fitter.fit(self.params, self.xs, self.ys)

    def compute_fitted(self):
        return self.shape_fitter.shape(self.xs, **self.params_dict)

    def interpolate(self, x):
        return self.shape_fitter.shape(x, **self.params_dict)

    def peak_shape_fit(self):
        xs, ys = self.xs, self.ys
        params = self.shape_fitter.guess(xs, ys)
        fit = leastsq(self.shape_fitter.fit,
                      params, (xs, ys))
        params = fit[0]
        self.params = params
        self.params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)

    def iterfits(self):
        yield self.compute_fitted()


def shape_fit_test(chromatogram, smooth=True):
    return ChromatogramShapeFitter(chromatogram, smooth).line_test


def peak_indices(x, min_height=0):
    """Find the index of local maxima.

    Parameters
    ----------
    x : np.ndarray
        Data to find local maxima in
    min_height : float, optional
        Minimum peak height

    Returns
    -------
    np.ndarray[int]
        Indices of maxima in x

    References
    ----------
    https://github.com/demotu/BMC/blob/master/functions/detect_peaks.py
    """
    if x.size < 3:
        return np.array([], dtype=int)
    # find indices of all peaks
    dx = x[1:] - x[:-1]
    rising_edges = np.where((np.hstack((dx, 0)) <= 0) &
                            (np.hstack((0, dx)) > 0))[0]
    falling_edges = np.where((np.hstack((dx, 0)) < 0) &
                             (np.hstack((0, dx)) >= 0))[0]
    indices = np.unique(np.hstack((rising_edges, falling_edges)))
    if indices.size and min_height > 0:
        indices = indices[x[indices] >= min_height]
    return indices


class MultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_peaks=5, smooth=True, fitter=BiGaussianModel()):
        super(MultimodalChromatogramShapeFitter, self).__init__(chromatogram, smooth=smooth, fitter=fitter)
        self.max_peaks = max_peaks
        self.params_list = []
        self.params_dict_list = []

        if self.is_invalid():
            self.handle_invalid()
        else:
            try:
                self.extract_arrays()
                self.peak_shape_fit()
                self.perform_line_test()
            except TypeError:
                self.handle_invalid()

    @property
    def fit_parameters(self):
        return self.params_dict_list

    def __repr__(self):
        return "MultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def peak_shape_fit(self):
        return self.set_up_peak_fit()

    def set_up_peak_fit(self, ys=None, min_height=0, peak_count=0):
        xs = self.xs
        if ys is None:
            ys = self.ys
        params = self.shape_fitter.guess(xs, ys)
        params_dict = self.shape_fitter.params_to_dict(params)

        indices = peak_indices(ys, min_height)
        center = xs[max(indices, key=lambda x: ys[x])]
        params_dict['center'] = center

        fit = leastsq(self.shape_fitter.fit,
                      list(params_dict.values()), (xs, ys))
        params = fit[0]
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        self.params_list.append(params)
        self.params_dict_list.append(params_dict)

        residuals = self.shape_fitter.fit(params, xs, ys)

        fitted_apex_index = search.get_nearest(xs, params_dict['center'], 0)
        fitted_apex = ys[fitted_apex_index]

        new_min_height = fitted_apex * 0.5

        if new_min_height < min_height:
            min_height *= 0.85
        else:
            min_height = new_min_height

        indices = peak_indices(residuals, min_height)

        peak_count += 1
        if indices.size and peak_count < self.max_peaks:
            residuals, params_dict = self.set_up_peak_fit(residuals, min_height, peak_count=peak_count)

        return residuals, params_dict

    def compute_fitted(self):
        xs = self.xs
        fitted = np.zeros_like(xs)
        for params_dict in self.params_dict_list:
            fitted += self.shape_fitter.shape(xs, **params_dict)
        return fitted

    def interpolate(self, x):
        fitted = np.zeros_like(x)
        for params_dict in self.params_dict_list:
            fitted += self.shape_fitter.shape(x, **params_dict)
        return fitted

    def compute_residuals(self):
        return self.ys - self.compute_fitted()

    def iterfits(self):
        xs = self.xs
        for params_dict in self.params_dict_list:
            yield self.shape_fitter.shape(xs, **params_dict)


class AdaptiveMultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_peaks=5, smooth=True, fitters=None):
        if fitters is None:
            fitters = (GaussianModel(), BiGaussianModel(), PenalizedSkewedGaussianModel(),)
        super(AdaptiveMultimodalChromatogramShapeFitter, self).__init__(
            chromatogram, smooth=smooth, fitter=fitters[0])
        self.max_peaks = max_peaks
        self.fitters = fitters
        self.params_list = []
        self.params_dict_list = []

        self.alternative_fits = []
        self.best_fit = None

        if self.is_invalid():
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()

    def is_invalid(self):
        return len(self.chromatogram) < MIN_POINTS

    @property
    def fit_parameters(self):
        return self.best_fit.fit_parameters

    @property
    def xs(self):
        try:
            return self.best_fit.xs
        except AttributeError:
            return self._xs

    @xs.setter
    def xs(self, value):
        self._xs = value

    @property
    def ys(self):
        try:
            return self.best_fit.ys
        except AttributeError:
            return self._ys

    @ys.setter
    def ys(self, value):
        self._ys = value

    def compute_fitted(self):
        return self.best_fit.compute_fitted()

    def interpolate(self, x):
        return self.best_fit.interpolate(x)

    def compute_residuals(self):
        return self.best_fit.compute_residuals()

    def _get_gap_size(self):
        return np.average(self.xs[1:] - self.xs[:-1],
                          weights=(self.ys[1:] + self.ys[:-1])) * 2

    def has_sparse_tails(self):
        gap = self._get_gap_size()
        partition = [False, False]
        if self.xs[1] - self.xs[0] > gap:
            partition[0] = True
        if self.xs[-1] - self.xs[-2] > gap:
            partition[1] = True
        return partition

    def generate_trimmed_chromatogram_slices(self):
        for tails in set(product(*zip(self.has_sparse_tails(), [False, False]))):
            if not tails[0] and not tails[1]:
                continue
            if tails[0]:
                slice_start = self.xs[1]
            else:
                slice_start = self.xs[0]
            if tails[1]:
                slice_end = self.xs[-2]
            else:
                slice_end = self.xs[-1]
            subset = self.chromatogram.slice(slice_start, slice_end)
            yield subset

    def peak_shape_fit(self):
        for fitter in self.fitters:
            model_fit = ProfileSplittingMultimodalChromatogramShapeFitter(
                self.chromatogram, self.max_peaks, self.smooth, fitter=fitter)
            self.alternative_fits.append(model_fit)
            model_fit = MultimodalChromatogramShapeFitter(
                self.chromatogram, self.max_peaks, self.smooth, fitter=fitter)
            self.alternative_fits.append(model_fit)
            for subset in self.generate_trimmed_chromatogram_slices():
                model_fit = ProfileSplittingMultimodalChromatogramShapeFitter(
                    subset, self.max_peaks,
                    self.smooth, fitter=fitter)
                self.alternative_fits.append(model_fit)
        ix = np.nanargmin([f.line_test for f in self.alternative_fits])
        # self.best_fit = min(self.alternative_fits, key=lambda x: x.line_test)
        self.best_fit = self.alternative_fits[ix]
        self.params_list = self.best_fit.params_list
        self.params_dict_list = self.best_fit.params_dict_list
        self.shape_fitter = self.best_fit.shape_fitter

    def perform_line_test(self):
        self.line_test = self.best_fit.line_test

    def plot(self, *args, **kwargs):
        return self.best_fit.plot(*args, **kwargs)

    def iterfits(self):
        xs = self.xs
        for params_dict in self.params_dict_list:
            yield self.shape_fitter.shape(xs, **params_dict)

    def __repr__(self):
        return "AdaptiveMultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)


class SplittingPoint(object):
    __slots__ = ["first_maximum", "minimum", "second_maximum", "minimum_index", "total_distance"]

    def __init__(self, first_maximum, minimum, second_maximum, minimum_index):
        self.first_maximum = first_maximum
        self.minimum = minimum
        self.second_maximum = second_maximum
        self.minimum_index = minimum_index
        self.total_distance = self.compute_distance()

    def compute_distance(self):
        return (self.first_maximum - self.minimum) + (self.second_maximum - self.minimum)

    def __repr__(self):
        return "SplittingPoint(%0.4f, %0.4f, %0.4f, %0.2f, %0.3e)" % (
            self.first_maximum, self.minimum, self.second_maximum, self.minimum_index, self.total_distance)


class ProfileSplittingMultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_splits=3, smooth=True, fitter=BiGaussianModel()):
        super(ProfileSplittingMultimodalChromatogramShapeFitter, self).__init__(
            chromatogram, smooth=smooth, fitter=fitter)
        self.max_splits = max_splits
        self.params_list = []
        self.params_dict_list = []
        self.partition_sites = []

        if self.is_invalid():
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()

    def __repr__(self):
        return "ProfileSplittingMultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def _extreme_indices(self, ys):
        maxima_indices = peak_indices(ys)
        minima_indices = peak_indices(-ys)
        return maxima_indices, minima_indices

    def locate_extrema(self, xs=None, ys=None):
        if xs is None:
            xs = self.xs
        if ys is None:
            ys = self.ys

        maxima_indices, minima_indices = self._extreme_indices(ys)
        candidates = []

        for i in range(len(maxima_indices)):
            max_i = maxima_indices[i]
            for j in range(i + 1, len(maxima_indices)):
                max_j = maxima_indices[j]
                for k in range(len(minima_indices)):
                    min_k = minima_indices[k]
                    y_i = ys[max_i]
                    y_j = ys[max_j]
                    y_k = ys[min_k]
                    if max_i < min_k < max_j and (y_i - y_k) > (y_i * 0.01) and (
                            y_j - y_k) > (y_j * 0.01):
                        point = SplittingPoint(y_i, y_k, y_j, xs[min_k])
                        candidates.append(point)
        if candidates:
            best_point = max(candidates, key=lambda x: x.total_distance)
            self.partition_sites.append(best_point)

        return candidates

    def build_partitions(self):
        segments = []

        last_x = self.xs.min() - 1
        for point in self.partition_sites:
            mask = (self.xs <= point.minimum_index) & (self.xs > last_x)
            if any(mask):
                xs, ys = self.xs[mask], self.ys[mask]
                # if len(xs) > 1:
                segments.append((xs, ys))
            last_x = point.minimum_index
        mask = self.xs > last_x
        if any(mask):
            xs, ys = self.xs[mask], self.ys[mask]
            # if len(xs) > 1:
            segments.append((xs, ys))
        return segments

    def set_up_peak_fit(self, xs, ys):
        params = self.shape_fitter.guess(xs, ys)
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        if len(params) > len(xs):
            self.params_list.append(params)
            self.params_dict_list.append(params_dict)
            return ys, params_dict

        fit = leastsq(self.shape_fitter.fit,
                      list(params_dict.values()), (xs, ys))
        params = fit[0]
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        self.params_list.append(params)
        self.params_dict_list.append(params_dict)

        residuals = self.shape_fitter.fit(params, xs, ys)
        return residuals, params_dict

    def peak_shape_fit(self):
        self.locate_extrema()
        for segment in self.build_partitions():
            self.set_up_peak_fit(*segment)

    def compute_fitted(self):
        fitted = []
        for segment, params_dict in zip(self.build_partitions(), self.params_dict_list):
            fitted.append(self.shape_fitter.shape(segment[0], **params_dict))
        return np.concatenate(fitted)

    def interpolate(self, x):
        fitted = np.zeros_like(x)
        for params_dict in self.params_dict_list:
            fitted += self.shape_fitter.shape(x, **params_dict)
        return fitted

    def compute_residuals(self):
        return self.ys - self.compute_fitted()

    def iterfits(self):
        for segment, params_dict in zip(self.build_partitions(), self.params_dict_list):
            yield self.shape_fitter.shape(segment[0], **params_dict)


try:
    _bigaussian_shape = bigaussian_shape
    _skewed_gaussian_shape = skewed_gaussian_shape
    _gaussian_shape = gaussian_shape

    from ms_deisotope._c.feature_map.shape_fitter import (
        bigaussian_shape, skewed_gaussian_shape, gaussian_shape)

    has_c = True
except ImportError:
    has_c = False

try:
    _plocate_extrema = ProfileSplittingMultimodalChromatogramShapeFitter.locate_extrema
    from glycresoft._c.scoring.shape_fitter import locate_extrema as _clocate_extrema
    ProfileSplittingMultimodalChromatogramShapeFitter.locate_extrema = _clocate_extrema
except ImportError:
    pass


class ChromatogramShapeModel(ScoringFeatureBase):
    feature_type = "line_score"

    def __init__(self, smooth=True):
        self.smooth = smooth

    def fit(self, chromatogram, *args, **kwargs):
        return AdaptiveMultimodalChromatogramShapeFitter(chromatogram, smooth=self.smooth)

    def score(self, chromatogram, *args, **kwargs):
        fit = self.fit(chromatogram, *args, **kwargs)
        return max(1 - fit.line_test, epsilon)



'''The barebone-essentials of weighted ordinary least squares and
a RANSAC-wrapper of it.
'''
import random
import functools

from typing import Tuple, Optional

import numpy as np
from scipy import stats

from glycresoft._c.scoring.elution_time_grouping import posprocess_linear_fit

def is_square(x: np.ndarray):
    if x.ndim == 2:
        return x.shape[0] == x.shape[1]
    return False


class WLSSolution(object):
    '''A structured container for :func:`weighted_linear_regression_fit`
    output.
    '''
    __slots__ = ('yhat', 'parameters', 'data', 'weights', 'residuals',
                 'projection_matrix', 'rss', 'press', 'R2', 'variance_matrix')

    yhat: np.ndarray
    parameters: np.ndarray
    data: Tuple[np.ndarray, np.ndarray]
    weights: np.ndarray
    residuals: np.ndarray
    projection_matrix: np.ndarray
    rss: float
    press: float
    R2: float
    variance_matrix: np.ndarray

    def __init__(self, yhat, parameters, data, weights, residuals, projection_matrix, rss, press, R2, variance_matrix):
        self.yhat = yhat
        self.parameters = parameters
        self.data = data
        self.weights = weights
        self.residuals = residuals
        self.projection_matrix = projection_matrix
        self.rss = rss
        self.press = press
        self.R2 = R2
        self.variance_matrix = variance_matrix

    @property
    def y(self) -> np.ndarray:
        return self.data[1]

    @property
    def X(self) -> np.ndarray:
        return self.data[0]

    def __getstate__(self):
        state = {}
        state['parameters'] = self.parameters
        state['data'] = self.data
        if is_square(self.weights):
            self.weights = np.diag(self.weights)
        state['weights'] = self.weights
        # state['projection_matrix'] = self.projection_matrix
        state['projection_matrix'] = None
        state['rss'] = self.rss
        state['press'] = self.press
        state['R2'] = self.R2
        state['variance_matrix'] = self.variance_matrix
        return state

    def __setstate__(self, state):
        X, y = state['data']
        self.parameters = state['parameters']
        self.data = state['data']
        self.weights = state['weights']
        if is_square(self.weights):
            self.weights = np.diag(self.weights)
        self.projection_matrix = state['projection_matrix']
        self.rss = state['rss']
        self.press = state['press']
        self.R2 = state['R2']
        self.variance_matrix = state['variance_matrix']
        self.yhat = X.dot(self.parameters)
        self.residuals = y - self.yhat

    def __repr__(self):
        template = "{self.__class__.__name__}({fields})"
        fields = ', '.join(["%s=%r" % (k, getattr(self, k)) for k in self.__slots__])
        return template.format(self=self, fields=fields)


SMALL_ERROR = 1e-5

T_ISF = functools.lru_cache(None)(stats.t.isf)


def prepare_arrays_for_linear_fit(x: np.ndarray, y: np.ndarray, w: Optional[np.ndarray]=None):
    """Prepare data for estimating parameter values using the
    weighted ordinary least squares method implemented in :func:`weighted_linear_regerssion_fit`

    Parameters
    ----------
    x : :class:`np.ndarray`
        The data vector or matrix of predictors. Does not contain a common
        intercept.
    y : :class:`np.ndarray`
        The response variable, should have the same outer dimension as x
    w : :class:`np.ndarray`, optional
        The optional weight matrix. If omitted, the identity matrix of appropriate
        shape will be returned.

    Returns
    -------
    X : :class:`np.ndarray`
        The predictors, with a common intercept term added.
    y : :class:`np.ndarray`
        The response variable.
    w : :class:`np.ndarray`
        The weight matrix of X
    """
    X = np.vstack((np.ones(len(x)), np.array(x))).T
    Y = np.array(y)
    if w is None:
        W = np.ones(Y.shape[0])
    else:
        W = np.array(w)
    return X, Y, W


def weighted_linear_regression_fit_ridge(X: np.ndarray, y: np.ndarray,
                                         w: Optional[np.ndarray]=None,
                                         alpha: Optional[np.ndarray]=None,
                                         prepare: bool=False) -> WLSSolution:
    if prepare:
        X, y, w = prepare_arrays_for_linear_fit(X, y, w)
    elif w is None:
        w = np.ones(y.shape[0])
    if alpha is None:
        alpha = 0.0
    elif isinstance(alpha, np.ndarray):
        if alpha.ndim == 1:
            alpha = np.diag(alpha)
    else:
        raise TypeError("Could not infer the format for alpha")

    if w.ndim == 1:
        Xtw = X.T * w
    else:
        Xtw = X.T @ w
        w = np.ascontiguousarray(np.diag(w))
    V = np.linalg.pinv((Xtw @ X) + alpha)
    A = V @ Xtw
    del Xtw
    B = A @ y
    H = X @ A
    yhat = X @ B
    residuals = (y - yhat)
    # leave_one_out_error = residuals / (1 - np.diag(H))
    # press = (np.diag(w) * leave_one_out_error * leave_one_out_error).sum()
    # rss = (np.diag(w) * residuals * residuals).sum()
    # tss = (y - y.mean())
    # tss = (np.diag(w) * tss * tss).sum()
    diag_H = np.require(np.ascontiguousarray(np.diag(H)), requirements="W")
    w = np.require(w, requirements="W")
    press, rss, tss = posprocess_linear_fit(y, residuals, w, diag_H)
    if tss == 0:
        tss = 1e-16
    return WLSSolution(
        yhat, B, (X, y), w, residuals, H,
        rss, press, 1 - (rss / (tss)), V)


def weighted_linear_regression_fit(X: np.ndarray, y: np.ndarray,
                                   w: Optional[np.ndarray]=None, prepare: bool=False) -> WLSSolution:
    """Fit a linear model using weighted least squares.

    Parameters
    ----------
    x : :class:`np.ndarray`
        The data vector or matrix of predictors
    y : :class:`np.ndarray`
        The response variable, should have the same outer dimension as x
    w : :class:`np.ndarray`, optional
        The optional weight matrix
    prepare : bool, optional
        Whether or not to pass the parameters through :func:`prepare_arrays_for_linear_fit`

    Returns
    -------
    WLSSolution
    """
    if prepare:
        X, y, w = prepare_arrays_for_linear_fit(X, y, w)
    elif w is None:
        w = np.eye(y.shape[0])
    if w.ndim == 1:
        Xtw = X.T * w
    else:
        Xtw = X.T @ w
        w = np.ascontiguousarray(np.diag(w))
    V = np.linalg.pinv(Xtw.dot(X))
    A = V.dot(Xtw)
    B = A.dot(y)
    H = X.dot(A)
    yhat = X.dot(B)
    residuals = (y - yhat)
    # leave_one_out_error = residuals / (1 - np.diag(H))
    # press = (np.diag(w) * leave_one_out_error * leave_one_out_error).sum()
    # rss = (np.diag(w) * residuals * residuals).sum()
    # tss = (y - y.mean())
    # tss = (np.diag(w) * tss * tss).sum()
    diag_H = np.require(np.ascontiguousarray(np.diag(H)), requirements='W')
    w = np.require(w, requirements='W')
    press, rss, tss = posprocess_linear_fit(y, residuals, w, diag_H)
    return WLSSolution(
        yhat, B, (X, y), w, residuals, H,
        rss, press, 1 - (rss / (tss)), V)


def ransac(x, y, w=None, max_trials=100, regularize_alpha=None):
    '''
    RANSAC Regression, inspired heavily by sklearn's
    much more complex implementation
    '''
    X = x
    residual_threshold = np.median(np.abs(y - np.median(y)))

    if w is None:
        w = np.eye(y)

    def loss(y_true, y_pred):
        return np.abs(y_true - y_pred)

    n_trials = 0
    n_samples = X.shape[0]
    min_samples = X.shape[1] * 5
    if min_samples > X.shape[0]:
        min_samples = X.shape[1] + 1

    if min_samples > X.shape[0]:
        if regularize_alpha is not None:
            return weighted_linear_regression_fit_ridge(X, y, w, regularize_alpha)
        return weighted_linear_regression_fit(X, y, w)

    sample_indices = np.arange(n_samples)

    rng = random.Random(1)

    n_inliers_best = 1
    score_best = -np.inf
    X_inlier_best = None
    y_inlier_best = None
    w_inlier_best = None

    while n_trials < max_trials:
        n_trials += 1
        subset_ix = rng.sample(sample_indices, min_samples)
        X_subset = X[subset_ix]
        y_subset = y[subset_ix]
        w_subset = np.diag(np.diag(w)[subset_ix])

        # fit parameters on random subset of the data
        if regularize_alpha is not None:
            fit = weighted_linear_regression_fit_ridge(
                X_subset, y_subset, w_subset, regularize_alpha)
        else:
            fit = weighted_linear_regression_fit(X_subset, y_subset, w_subset)

        # compute goodness of fit for the fitted parameters with
        # the full dataset
        yhat = np.dot(X, fit.parameters)
        residuals_subset = loss(y, yhat)

        # locate inliers based on residual threshold
        inlier_subset_mask = residuals_subset < residual_threshold
        n_inliers_subset = inlier_subset_mask.sum()

        # determine the quality of the fitted parameters for
        # the inliers using R2
        inlier_subset_ix = sample_indices[inlier_subset_mask]
        X_inlier_subset = X[inlier_subset_ix]
        y_inlier_subset = y[inlier_subset_ix]
        w_inlier_subset = np.diag(np.diag(w)[inlier_subset_ix])
        # w_inlier_best = 1

        yhat_inlier_subset = X_inlier_subset.dot(fit.parameters)
        rss = (w_inlier_subset * np.square(
            y_inlier_subset - yhat_inlier_subset)).sum()
        tss = (w_inlier_subset * np.square(
            y_inlier_subset - y_inlier_subset.mean())).sum()

        score_subset = 1 - (rss / tss)

        # If the number of inliers chosen hasn't improved and the score hasn't
        # improved, don't update the current best
        if n_inliers_subset < n_inliers_best and score_subset < score_best:
            continue

        score_best = score_subset
        X_inlier_best = X_inlier_subset
        y_inlier_best = y_inlier_subset
        w_inlier_best = w_inlier_subset

    if regularize_alpha is None:
        return weighted_linear_regression_fit_ridge(
            X_inlier_best, y_inlier_best, w_inlier_best, regularize_alpha)
    # fit the final best inlier set for the final parameters
    return weighted_linear_regression_fit(
        X_inlier_best, y_inlier_best, w_inlier_best)


def fitted_interval(solution: WLSSolution, x0: np.ndarray,
                    y0: np.ndarray, alpha: float=0.05) -> np.ndarray:
    n = len(solution.residuals)
    k = len(solution.parameters)
    df = n - k
    sigma2 = solution.rss / df
    xtx_inv = solution.variance_matrix
    h = x0.dot(xtx_inv).dot(x0.T)
    if not np.isscalar(h):
        h = np.diag(h)

    error_of_fit = np.sqrt(sigma2 * h)

    t = T_ISF(alpha / 2., df)
    width = t * error_of_fit
    return np.stack([y0 - width, y0 + width])


def prediction_interval(solution, x0, y0, alpha=0.05):
    """Calculate the prediction interval around `x0` with response
    `y0` given the `solution`.

    Parameters
    ----------
    solution : :class:`WLSSolution`
        The fitted model
    x0 : :class:`np.ndarray`
        The new predictors
    y0 : float
        The predicted response
    alpha : float, optional
        The prediction interval width. Defaults to 0.05

    Returns
    -------
    :class:`np.ndarray` :
        The lower and upper bound of the prediction interval.
    """
    n = len(solution.residuals)
    k = len(solution.parameters)
    xtx_inv = solution.variance_matrix
    df = n - k
    sigma2 = solution.rss / df
    h = x0.dot(xtx_inv).dot(x0.T)
    if not np.isscalar(h):
        h = np.diag(h)
    error_of_prediction = np.sqrt(sigma2 * (1 + h))

    t = T_ISF(alpha / 2., df)
    width = t * error_of_prediction
    return np.stack([y0 - width, y0 + width])



from glycresoft.scoring import (
    DummyFeature,
    MassScalingMassShiftScoringModel)

from glycresoft.scoring.base import (
    degree_of_sialylation,
    CompositionDispatchingModel)

from .utils import make_model_loader
from .features import register_feature


load_model = make_model_loader(MassScalingMassShiftScoringModel)

AmmoniumMassShiftFeature = load_model("ammonium_adduct_model")
MethylLossFeature = load_model("methyl_loss_adduct_model")


register_feature("permethylated_ammonium_adducts", AmmoniumMassShiftFeature)
register_feature("methyl_loss", MethylLossFeature)


class PermethylatedAmmoniumAndMethylLossModel(DummyFeature):
    def __init__(self, boosting=True):
        DummyFeature.__init__(
            self, name=self.__class__.__name__, feature_type="mass_shift_score")
        self.ammonium = AmmoniumMassShiftFeature
        self.methyl_loss = MethylLossFeature
        self.boosting = boosting

    def score(self, chromatogram, *args, **kwargs):
        methyl_score = self.methyl_loss.score(chromatogram)
        # the methyl loss scoring model is < 0.2 if only a loss
        # was matched, 0.5 if only unmodified was matched, and
        # > 0.5 if both unmodified and a loss was matched.
        #
        # to make this scorer pass through directly to the ammonium
        # model when there was no methyl loss detected, either the
        # case where unmodified is detecteded should map to essentially
        # 1.0 or the case where no loss is registered
        methyl_score *= 2
        if not self.boosting:
            if methyl_score > 1:
                methyl_score = 1
        # there was no methyl loss
        if methyl_score < 1e-3:
            methyl_score = 1
        # elif methyl_score > 1:
        #     methyl_score = 1
        score = methyl_score * self.ammonium.score(chromatogram)
        if score > 1:
            score = 1 - 1e-3
        return score


PermethylatedAmmoniumAndMethylLossFeature = PermethylatedAmmoniumAndMethylLossModel()

register_feature(
    "permethylated_ammonium_adducts_methyl_loss",
    PermethylatedAmmoniumAndMethylLossFeature)


AsialoFormateMassShiftFeature = load_model("asialo_formate_adduct_model")
MonosialoFormateMassShiftFeature = load_model("monosialo_formate_adduct_model")
DisialoFormateMassShiftFeature = load_model("disialo_formate_adduct_model")
TrisialoFormateMassShiftFeature = load_model("trisialo_formate_adduct_model")
TetrasialoFormateMassShiftFeature = load_model("tetrasialo_formate_adduct_model")

_GeneralizedFormateMassShiftModelTable = CompositionDispatchingModel({
    lambda x: degree_of_sialylation(x) == 0: AsialoFormateMassShiftFeature,
    lambda x: degree_of_sialylation(x) == 1: MonosialoFormateMassShiftFeature,
    lambda x: degree_of_sialylation(x) == 2: DisialoFormateMassShiftFeature,
    lambda x: degree_of_sialylation(x) == 3: TrisialoFormateMassShiftFeature,
    lambda x: degree_of_sialylation(x) > 3: TetrasialoFormateMassShiftFeature
}, AsialoFormateMassShiftFeature)


class GeneralizedFormateMassShiftModel(DummyFeature):
    def __init__(self):
        DummyFeature.__init__(
            self, name=self.__class__.__name__, feature_type="mass_shift_score")
        self.table = _GeneralizedFormateMassShiftModelTable

    def score(self, chromatogram, *args, **kwargs):
        score = self.table.score(chromatogram, *args, **kwargs)
        # This feature can easily be 1.0 under certain conditions, which
        # leads to undesirable score inflation. The maximum contribution
        # this should have is capped to 0.7 (logit(0.7) = 0.8472) which
        # should not lead to many more assignments being retained while
        # still penalizing inappropriate formate adduction states
        return min(score, 0.7)


# pickle loading alias
GeneralizedFormateAdductModel = GeneralizedFormateMassShiftModel

GeneralizedFormateMassShiftFeature = GeneralizedFormateMassShiftModel()


register_feature("formate_adduct_model", GeneralizedFormateMassShiftFeature)



class LRUNode(object):

    def __init__(self, data, forward, backward):
        self.data = data
        self.forward = forward
        self.backward = backward

    def __hash__(self):
        return hash(self.data)

    def __eq__(self, other):
        return self.data == other.data

    def __repr__(self):
        return "LRUNode(%s)" % self.data


class LRUCache(object):

    def __init__(self):
        self.head = LRUNode(None, None, None)
        self.head.forward = self.head
        self.head.backward = self.head
        self._mapping = dict()

    def move_node_up(self, node):
        was_head_lru = self.head.backward is self.head
        if was_head_lru:
            raise ValueError("Head cannot be LRU")
        mover_forward = node.forward
        mover_backward = node.backward

        pushed = self.head.forward
        # If the most-recently used node is being moved up,
        # short-circuit
        if pushed == node:
            return
        # If the node being moved up happens to form a cycle with
        # the head node, (which should technically be handled by the
        # above short-circuit) don't make any changes as this will
        # lead to a detached head node.
        if mover_backward is self.head and mover_forward is self.head:
            return
        mover_backward.forward = mover_forward
        mover_forward.backward = mover_backward

        pushed.backward = node

        node.backward = self.head
        node.forward = pushed

        self.head.forward = node
        now_head_lru = self.head.backward is self.head
        if now_head_lru:
            raise ValueError("Head became LRU!")

    def add_node(self, data):
        if data in self._mapping:
            self.hit_node(data)
            return
        node = LRUNode(data, None, None)

        pushed = self.head.forward
        self.head.forward = node
        node.backward = self.head
        node.forward = pushed
        pushed.backward = node

        self._mapping[node.data] = node

    def get_least_recently_used(self):
        lru = self.head.backward
        if lru is self.head:
            raise ValueError("Head node cannot be LRU!")
        return lru.data

    def hit_node(self, k):
        out = self._mapping[k]
        self.move_node_up(out)

    def unspool(self):
        chain = []
        current = self.head
        while current.forward is not self.head:
            chain.append(current)
            current = current.forward
        chain.append(current)
        return chain

    def remove_node(self, data):
        # print("Removing ", data, id(data))
        node = self._mapping[data]
        fwd = node.forward
        # assert fwd is not node
        bck = node.backward
        # assert bck is not node
        fwd.backward = bck
        bck.forward = fwd
        self._mapping.pop(data)
        # assert node not in self.unspool()[1:]
        # print("Removed")

    def clear(self):
        self.head = LRUNode(None, None, None)
        self.head.forward = self.head
        self.head.backward = self.head
        self._mapping.clear()


try:
    from glycresoft._c.structure.lru import LRUCache, LRUNode
except ImportError:
    pass


class LRUMapping(object):
    def __init__(self, max_size=512):
        self.max_size = max_size
        self.store = dict()
        self.lru = LRUCache()

    def __getitem__(self, key):
        value = self.store[key]
        self.lru.hit_node(key)
        return value

    def __setitem__(self, key, value):
        self.lru.add_node(key)
        self.store[key] = value
        self._check_size()

    def __delitem__(self, key):
        del self.store[key]
        self.lru.remove_node(key)

    def _check_size(self):
        n = len(self.store)
        while n > self.max_size:
            key = self.lru.get_least_recently_used()
            self.store.pop(key)
            self.lru.remove_node(key)
            n -= 1

    def __contains__(self, key):
        contained = key in self.store
        return contained

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def get(self, key, default=None):
        try:
            value = self[key]
            return value
        except KeyError:
            return default

    def pop(self, key, default=None):
        try:
            value = self[key]
            del self[key]
            return value
        except KeyError:
            return default

    def clear(self):
        self.store.clear()
        self.lru.clear()


try:
    from glycresoft._c.structure.lru import LRUMapping
except ImportError:
    pass



from glycopeptidepy.structure.fragment import ChemicalShift, IonSeries, SimpleFragment, Composition
from glycopeptidepy.structure.fragmentation_strategy import EXDFragmentationStrategy, HCDFragmentationStrategy

from ...spectrum_match import SpectrumMatcherBase


class PeptideSpectrumMatcherBase(SpectrumMatcherBase):
    def get_fragments(self, series, strategy=None, include_neutral_losses=False):
        fragments = self.target.get_fragments(
            series, strategy=strategy, include_neutral_losses=include_neutral_losses)
        return fragments

    def _match_backbone_series(self, series, error_tolerance=2e-5, masked_peaks=None, strategy=None,
                               include_neutral_losses=False):
        if strategy is None:
            strategy = HCDFragmentationStrategy
        if masked_peaks is None:
            masked_peaks = set()
        for frags in self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses):
            for frag in frags:
                for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                    if peak.index.neutral_mass in masked_peaks:
                        continue
                    self.solution_map.add(peak, frag)

    def _theoretical_mass(self):
        return self.target.total_mass

    def _match_precursor(self, error_tolerance=2e-5, masked_peaks=None, include_neutral_losses=False):
        if masked_peaks is None:
            masked_peaks = set()
        mass = self.target.total_mass
        frag = SimpleFragment("M", mass, IonSeries.precursor, None)
        for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
            if peak.index.neutral_mass in masked_peaks:
                continue
            masked_peaks.add(peak)
            self.solution_map.add(peak, frag)
        if include_neutral_losses:
            delta = -Composition("NH3")
            for peak in self.spectrum.all_peaks_for(frag.mass + delta.mass, error_tolerance):
                if peak.index.neutral_mass in masked_peaks:
                    continue
                masked_peaks.add(peak)
                shifted_frag = frag.clone()
                shifted_frag.set_chemical_shift(ChemicalShift("-NH3", delta))
                self.solution_map.add(peak, shifted_frag)
            delta = -Composition("H2O")
            for peak in self.spectrum.all_peaks_for(frag.mass + delta.mass, error_tolerance):
                if peak.index.neutral_mass in masked_peaks:
                    continue
                masked_peaks.add(peak)
                shifted_frag = frag.clone()
                shifted_frag.set_chemical_shift(ChemicalShift("-H2O", delta))
                self.solution_map.add(peak, shifted_frag)

    def match(self, error_tolerance=2e-5, *args, **kwargs):
        masked_peaks = set()
        include_neutral_losses = kwargs.get("include_neutral_losses", False)
        is_hcd = self.is_hcd()
        is_exd = self.is_exd()
        if not is_hcd and not is_exd:
            is_hcd = True

        self._match_precursor(error_tolerance, masked_peaks, include_neutral_losses)

        # handle N-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, HCDFragmentationStrategy,
                include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)
            self._match_backbone_series(
                IonSeries.c, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)

        # handle C-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, HCDFragmentationStrategy,
                include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)
            self._match_backbone_series(
                IonSeries.z, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)
        return self


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import (
        PeptideSpectrumMatcherBase_match_backbone_series, _match_precursor)
    PeptideSpectrumMatcherBase._match_backbone_series = PeptideSpectrumMatcherBase_match_backbone_series
    PeptideSpectrumMatcherBase._match_precursor = _match_precursor
except ImportError:
    pass



from glypy.utils import Enum


class SentinelToken(object):
    """An object to hold opaque identity information regarding a worker process,
    to be used to signal that a worker process has received a signal

    Attributes
    ----------
    token : object
        The opaque identity
    """

    def __init__(self, token):
        self.token = token

    def __hash__(self):
        return hash(self.token)

    def __eq__(self, other):
        return self.token == other.token

    def __repr__(self):
        return "{self.__class__.__name__}({self.token})".format(self=self)


class ProcessDispatcherState(Enum):
    start = 1
    spawning = 2
    running = 3
    running_local_workers_live = 4
    running_local_workers_dead = 5
    terminating = 6
    terminating_workers_live = 7
    done = 8



import os
import warnings

from typing import (
    IO, Any, Dict, DefaultDict,
    Iterable, Iterator, List,
    Set, Tuple, Optional,
    Union, Hashable, NamedTuple
)
from collections import defaultdict

from ms_deisotope.data_source import ProcessedScan

from glycresoft.chromatogram_tree import Unmodified

from .spectrum_graph import ScanGraph


HitID = Hashable
GroupID = Hashable
DBHit = Any

WorkloadT = Union['WorkloadManager', 'WorkloadBatch']

# This number is overriden in spectrum_evaluation
# DEFAULT_WORKLOAD_MAX = 25e6
DEFAULT_WORKLOAD_MAX = -1
DEFAULT_WORKLOAD_MAX = float(os.environ.get("GLYCRESOFTDEFAULTWORKLOADMAX", DEFAULT_WORKLOAD_MAX))


class WorkloadManager(object):
    """An object that tracks the co-dependence of multiple query MSn
    scans on the same hit structures.

    By only dealing with a part of the entire workload at a time, the
    database search engine can avoid holding very large numbers of temporary
    objects in memory while waiting for related computations finish in
    random order.

    Attributes
    ----------
    hit_map : dict
        hit structure id to the structure object
    hit_to_scan_map : defaultdict(list)
        Maps hit structure id to a list of scan ids it matched
    scan_hit_type_map : defaultdict(str)
        Maps a scan id, hit structure id pair to the name of the mass
        shift type used to complete the match
    scan_map : dict
        Maps scan id to :class:`ms_deisotope.ProcessedScan` object
    scan_to_hit_map : defaultdict(list)
        Maps scan id to the structures it hit
    """

    scan_map: Dict[str, ProcessedScan]
    hit_map: Dict[HitID, DBHit]
    hit_to_scan_map: DefaultDict[HitID, List[str]]
    scan_to_hit_map: DefaultDict[str, List[HitID]]
    scan_hit_type_map: DefaultDict[Tuple[str, HitID], str]
    hit_group_map: DefaultDict[GroupID, Set[HitID]]

    _scan_graph: Optional[ScanGraph]

    def __init__(self):
        self.scan_map = dict()
        self.hit_map = dict()
        self.hit_to_scan_map = defaultdict(list)
        self.scan_to_hit_map = defaultdict(list)
        self.scan_hit_type_map = defaultdict(lambda: Unmodified.name)
        self._scan_graph = None
        self.hit_group_map = defaultdict(set)

    def clear(self):
        self.scan_map.clear()
        self.hit_map.clear()
        self.hit_to_scan_map.clear()
        self.scan_to_hit_map.clear()
        self.scan_hit_type_map.clear()
        self.hit_group_map.clear()
        self._scan_graph = None

    def pack(self):
        """
        Drop any scans that haven't been matched to a database hit.

        Returns
        -------
        WorkloadManager
        """
        unmatched_scans = set(self.scan_map) - set(self.scan_to_hit_map)
        for k in unmatched_scans:
            self.scan_map.pop(k, None)
        return self

    def update(self, other: 'WorkloadManager'):
        self.scan_map.update(other.scan_map)
        self.hit_map.update(other.hit_map)
        self.merge_hit_to_scan_map(other)
        self.merge_scan_to_hit_map(other)
        self.merge_hit_group_map(other)
        self.scan_hit_type_map.update(other.scan_hit_type_map)
        self._scan_graph = None

    def merge_hit_to_scan_map(self, other: 'WorkloadManager'):
        """Merge the :attr:`hit_to_scan_map` of another :class:`WorkloadManager`
        object into this one, adding non-redundant pairs.

        Parameters
        ----------
        other : :class:`WorkloadManager`
            The other workload manager to merge in.
        """
        for hit_id, scans in other.hit_to_scan_map.items():
            already_present = {scan.id for scan in self.hit_to_scan_map[hit_id]}
            for scan in scans:
                if scan.id not in already_present:
                    self.hit_to_scan_map[hit_id].append(scan)
                    self.scan_hit_type_map[scan.id, hit_id] = other.scan_hit_type_map[scan.id, hit_id]

    def merge_scan_to_hit_map(self, other: 'WorkloadManager'):
        """Merge the :attr:`scan_to_hit_map` of another :class:`WorkloadManager`
        object into this one, adding non-redundant pairs.

        Parameters
        ----------
        other : :class:`WorkloadManager`
            The other workload manager to merge in.
        """
        for scan_id, hit_ids in other.scan_to_hit_map.items():
            already_present = set(self.scan_to_hit_map[scan_id])
            for hit_id in hit_ids:
                if hit_id not in already_present:
                    self.scan_to_hit_map[scan_id].append(hit_id)
                    self.scan_hit_type_map[scan_id, hit_id] = other.scan_hit_type_map[scan_id, hit_id]

    def merge_hit_group_map(self, other: 'WorkloadManager'):
        for group_key, members in other.hit_group_map.items():
            self.hit_group_map[group_key].update(members)

    def add_scan(self, scan: ProcessedScan):
        """Register a Scan-like object for tracking

        Parameters
        ----------
        scan : Scan-like
        """
        self.scan_map[scan.id] = scan

    def add_scan_hit(self, scan: ProcessedScan, hit: DBHit, hit_type: str=Unmodified.name):
        """Register a matching relationship between ``scan`` and ``hit``
        with a mass shift of type ``hit_type``

        Parameters
        ----------
        scan : Scan-like
            The scan that was matched
        hit : StructureRecord-like
            The structure that was matched
        hit_type : str, optional
            The name of the :class:`.MassShift` that was used to
            bridge the total mass of the match.
        """
        self.hit_map[hit.id] = hit
        self.hit_to_scan_map[hit.id].append(scan)
        self.scan_to_hit_map[scan.id].append(hit.id)
        self.scan_hit_type_map[scan.id, hit.id] = hit_type

    def build_scan_graph(self, recompute=False) -> ScanGraph:
        """
        Constructs a graph of scans where scans
        with common hits are conneted by an edge.

        This can be used to determine the set of scans
        which all depend upon the same structures and
        so should be processed together to minimize the
        amount of work done to recompute fragments.

        Parameters
        ----------
        recompute : bool, optional
            If recompute is True, the graph will be built
            from scratch, ignoring the cached copy if it
            exists.

        Returns
        -------
        ScanGraph
        """
        if self._scan_graph is not None and not recompute:
            return self._scan_graph
        graph = ScanGraph(self.scan_map.values())
        for key, values in self.hit_to_scan_map.items():
            values = sorted(values, key=lambda x: x.index)
            for i, scan in enumerate(values):
                for other in values[i + 1:]:
                    graph.add_edge_between(scan.id, other.id, key)
        self._scan_graph = graph
        return graph

    def compute_workloads(self) -> List[Tuple]:
        """
        Determine how much work is needed to evaluate all matches
        to a set of common scans.

        Construct a dependency graph relating scans and hit
        structures using :meth:`build_scan_graph` and build connected components.

        Returns
        -------
        list
            A list of tuples, each tuple containing a connected component and
            the number of comparisons contained in that component.
        """
        graph = self.build_scan_graph()
        components = graph.connected_components()
        workloads = []
        for m in components:
            workloads.append((m, sum([len(c.edges) for c in m])))
        return workloads

    def total_work_required(self, workloads=None):
        """
        Compute the total amount of work required to
        resolve this complete workload

        Returns
        -------
        int
            The amount of work contained in this load, or 1
            if the workload is empty.
        """
        return sum(map(len, self.scan_to_hit_map.values()))

    def log_workloads(self, handle: IO[str], workloads: Optional[List['WorkloadBatch']]=None):
        """
        Writes the current workload graph cluster sizes to a text file
        for diagnostic purposes

        Parameters
        ----------
        handle : file-like
            The file to write the workload graph cluster sizes to
        workloads : list, optional
            The precomputed workload graph clusters. If None, they will
            be computed with :meth:`compute_workloads`
        """
        if workloads is None:
            workloads = self.compute_workloads()
        workloads = sorted(workloads, key=lambda x: x[0][0].precursor_information.neutral_mass,
                           reverse=True)
        for scans, load in workloads:
            handle.write("Work: %d Comparisons\n" % (load,))
            for scan_node in scans:
                handle.write("%s\t%f\t%d\n" % (
                    scan_node.id,
                    scan_node.precursor_information.neutral_mass,
                    len(scan_node.edges)))

    def __iter__(self):
        yield self.total_size
        yield self.scan_map
        yield self.hit_map
        yield self.hit_to_scan_map
        yield self.scan_hit_type_map
        yield self.hit_group_map

    def __eq__(self, other: WorkloadT):
        if self.hit_map != other.hit_map:
            return False
        elif self.scan_map != other.scan_map:
            return False
        elif self.scan_hit_type_map != other.scan_hit_type_map:
            return False
        elif self.hit_to_scan_map != other.hit_to_scan_map:
            return False
        elif self.hit_group_map != other.hit_group_map:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    @property
    def total_size(self):
        """Compute the total amount of work required to
        resolve this complete workload

        Returns
        -------
        int
            The amount of work contained in this load, or 1
            if the workload is empty.
        """
        return sum(map(len, self.scan_to_hit_map.values()))

    @property
    def scan_count(self):
        """Return the number of scans in this workload

        Returns
        -------
        int
        """
        return len(self.scan_map)

    @property
    def hit_count(self):
        """Return the number of structures in this workload

        Returns
        -------
        int
        """
        return len(self.hit_map)

    def __repr__(self):
        template = 'WorkloadManager(scan_map_size={}, hit_map_size={}, total_cross_product={})'
        rendered = template.format(
            len(self.scan_map), len(self.hit_map),
            self.total_size)
        return rendered

    def batches(self, max_size: Optional[int]=None, ignore_groups: bool=False) -> Iterator['WorkloadBatch']:
        """Partition the workload into batches of approximately
        ``max_size`` comparisons.

        This guarantee is only approximate because a batch will
        only be broken at the level of a scan.

        Parameters
        ----------
        max_size : float, optional
            The maximum size of a workload

        Yields
        ------
        WorkloadBatch
            self-contained work required for a set of related scans
            whose total work is approximately ``max_size``
        """
        if max_size is None:
            max_size = DEFAULT_WORKLOAD_MAX
        elif max_size <= 0:
            max_size = float('inf')

        if max_size == float('inf'):
            current_scan_map = self.scan_map.copy()
            current_hit_map = self.hit_map.copy()
            current_hit_to_scan_map = {k: [v.id for v in vs] for k, vs in self.hit_to_scan_map.items()}
            current_scan_hit_type_map = self.scan_hit_type_map.copy()
            current_hit_group_map = self.hit_group_map.copy()
            batch = WorkloadBatch(
                self.total_size, current_scan_map,
                current_hit_map, current_hit_to_scan_map,
                current_scan_hit_type_map, current_hit_group_map)
            if batch.batch_size:
                yield batch
            return

        current_batch_size = 0
        current_scan_map = dict()
        current_hit_map = dict()
        current_hit_to_scan_map = defaultdict(list)
        current_scan_hit_type_map = defaultdict(lambda: Unmodified.name)
        current_hit_group_map = defaultdict(set)

        if self.hit_group_map and not ignore_groups:
            source = sorted(self.hit_group_map.items())
            batch_index = 0
            for group_id, hit_ids in source:
                current_hit_group_map[group_id] = hit_ids
                for hit_id in hit_ids:
                    current_hit_map[hit_id] = self.hit_map[hit_id]
                    for scan in self.hit_to_scan_map[hit_id]:
                        scan_id = scan.id
                        current_scan_map[scan_id] = self.scan_map[scan_id]
                        current_hit_to_scan_map[hit_id].append(scan_id)
                        current_scan_hit_type_map[
                            scan_id, hit_id] = self.scan_hit_type_map[scan_id, hit_id]
                        current_batch_size += 1

                if current_batch_size > max_size:
                    batch = WorkloadBatch(
                        current_batch_size, current_scan_map,
                        current_hit_map, current_hit_to_scan_map,
                        current_scan_hit_type_map, current_hit_group_map)
                    if batch.batch_size / max_size > 2:
                        warnings.warn("Batch %d has size %d, %0.3f%% larger than threshold" % (
                            batch_index, batch.batch_size, batch.batch_size * 100. / max_size))
                    batch_index += 1
                    current_batch_size = 0
                    current_scan_map = dict()
                    current_hit_map = dict()
                    current_hit_to_scan_map = defaultdict(list)
                    current_scan_hit_type_map = defaultdict(
                        lambda: Unmodified.name)
                    current_hit_group_map = defaultdict(set)
                    if batch.batch_size:
                        yield batch
        else:
            source = sorted(
                self.scan_map.items(),
                key=lambda x: x[1].precursor_information.neutral_mass)
            batch_index = 0
            for scan_id, scan in source:
                current_scan_map[scan_id] = scan
                for hit_id in self.scan_to_hit_map[scan_id]:
                    current_hit_map[hit_id] = self.hit_map[hit_id]
                    current_hit_to_scan_map[hit_id].append(scan_id)
                    current_scan_hit_type_map[
                        scan_id, hit_id] = self.scan_hit_type_map[scan_id, hit_id]
                    current_batch_size += 1

                if current_batch_size > max_size:
                    batch = WorkloadBatch(
                        current_batch_size, current_scan_map,
                        current_hit_map, current_hit_to_scan_map,
                        current_scan_hit_type_map, current_hit_group_map)
                    if batch.batch_size / max_size > 2:
                        warnings.warn("Batch %d has size %d, %0.3f%% larger than threshold" % (
                            batch_index, batch.batch_size, batch.batch_size * 100. / max_size))
                    batch_index += 1
                    current_batch_size = 0
                    current_scan_map = dict()
                    current_hit_map = dict()
                    current_hit_to_scan_map = defaultdict(list)
                    current_scan_hit_type_map = defaultdict(lambda: Unmodified.name)
                    current_hit_group_map = defaultdict(set)
                    if batch.batch_size:
                        yield batch

        if current_batch_size > 0:
            batch = WorkloadBatch(
                current_batch_size, current_scan_map,
                current_hit_map, current_hit_to_scan_map,
                current_scan_hit_type_map, current_hit_group_map)
            if batch.batch_size:
                yield batch

    @classmethod
    def merge(cls, workloads: Iterable[WorkloadT]):
        total = cls()
        for inst in workloads:
            total.update(inst)
        return total

    def mass_range(self) -> Tuple[float, float]:
        lo = float('inf')
        hi = 0
        for scan in self.scan_map.values():
            mass = scan.precursor_information.neutral_mass
            if mass < lo:
                lo = mass
            if mass > hi:
                hi = mass
        return (lo, hi)


class WorkloadBatch(NamedTuple):

    batch_size: int
    scan_map: Dict[str, ProcessedScan]
    hit_map: Dict[HitID, DBHit]
    hit_to_scan_map: DefaultDict[HitID, List[str]]
    scan_hit_type_map: DefaultDict[Tuple[str, HitID], str]
    hit_group_map: DefaultDict[GroupID, Set[HitID]]

    def clear(self):
        self.scan_map.clear()
        self.hit_map.clear()
        self.hit_to_scan_map.clear()
        self.scan_hit_type_map.clear()
        self.hit_group_map.clear()



from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg


def figax(*args, **kwargs):
    fig = Figure(*args, **kwargs)
    canvas = FigureCanvasAgg(fig)
    return fig.add_subplot(1, 1, 1)


def figure(*args, **kwargs):
    fig = Figure(*args, **kwargs)
    canvas = FigureCanvasAgg(fig)
    return fig
